# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import sys
import logging
import time
import tensorflow as tf
import numpy as np
tf.compat.v1.enable_v2_behavior()

# BARK imports
from bark.runtime.commons.parameters import ParameterServer

# tf agent imports
from tf_agents.drivers import dynamic_step_driver
from tf_agents.drivers import dynamic_episode_driver
from tf_agents.metrics import tf_metrics
from tf_agents.eval import metric_utils
from tf_agents.utils import common
from tf_agents.trajectories import time_step as ts

# BARK-ML imports
from bark_ml.library_wrappers.lib_tf_agents.runners.tfa_runner import TFARunner


class SACRunner(TFARunner):
  def __init__(self,
               environment=None,
               agent=None,
               params=None):
    TFARunner.__init__(self,
                       environment=environment,
                       agent=agent,
                       params=params)
    
    self._number_of_collections =\
      self._params["ML"]["SACRunner"]["NumberOfCollections", "", 40000]
    self._evaluation_interval =\
      self._params["ML"]["SACRunner"]["EvaluateEveryNSteps", "", 100] 

  def _train(self):
    iterator = iter(self._agent._dataset)
    iteration_start_time = time.time()
    global_iteration = self._agent._agent._train_step_counter.numpy()

    for i in range(int(global_iteration), self._number_of_collections):
      tf.keras.backend.clear_session()
      global_iteration = self._agent._agent._train_step_counter.numpy()
      tf.summary.experimental.set_step(global_iteration)

      t0 = time.time()
      self._collection_driver.run()
      self._log_collection_duration(start_time=t0, iteration=global_iteration)
 
      experience, _ = next(iterator)

      t0 = time.time()      
      self._agent._agent.train(experience)
      self._log_training_duration(start_time=t0, iteration=global_iteration)
            
      if global_iteration % self._evaluation_interval == 0:
        self._log_evaluation_interval_duration(
          start_time=iteration_start_time, 
          iteration=global_iteration)
        iteration_start_time = time.time()
        self._tracer.Reset()
        self.Run(
          num_episodes=self._params["ML"]["TFARunner"]["EvaluationSteps", "", 20],
          mode="training")
        self._agent.Save()

  def _log_collection_duration(self, start_time, iteration):
    with tf.name_scope("Durations"):
      tf.summary.scalar(
        "episode_collection_duration", time.time() - start_time, iteration)

  def _log_training_duration(self, start_time, iteration):
    with tf.name_scope("Durations"):
      tf.summary.scalar("training_duration", time.time() - start_time, iteration)

  def _log_evaluation_interval_duration(self, start_time, iteration):
    if iteration == 0: 
      return

    iterations = f'{iteration-self._evaluation_interval}-{iteration}'

    total_duration = time.time() - start_time
    mean_step_duration = total_duration / self._evaluation_interval
    self._logger.info(
      f'Training iterations {iterations} took {total_duration:.3f} seconds' +
      f' (avg. {mean_step_duration:.3f}s / iteration).')

    with tf.name_scope("Durations"):
      tf.summary.scalar("mean_step_duration",
                        mean_step_duration,
                        step=iteration)


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.


import numpy as np
import math

from bark.runtime.scenario.scenario_generation.config_readers.config_readers_interfaces import ConfigReaderControlledAgents

from bark.core.world.goal_definition import GoalDefinition, GoalDefinitionPolygon, GoalDefinitionStateLimits
from bark.runtime.commons.parameters import ParameterServer

# no one (in this road corridor) is a controlled agent
class NoneControlled(ConfigReaderControlledAgents):
   # returns list of size num agents with true or false depending if agent is controlled or not for each agent based on property, default_params_dict
  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    return [False] * len(agent_states), {}, config_param_object

# Select one agent randomly as controlled agent
class RandomSingleAgent(ConfigReaderControlledAgents):
  # returns list of size num agents with true or false depending if agent is controlled or not for each agent based on property, default_params_dict
  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    controlled_agent_idx = self.random_state.randint(low=0, high=len(agent_states), size=None) 
    controlled_list = [False] * len(agent_states)
    controlled_list[controlled_agent_idx] = True
    return controlled_list, {}, config_param_object


# Select one agent randomly as controlled agent
class PositioningSingleAgent(ConfigReaderControlledAgents):
  # returns list of size num agents with true or false depending if agent is controlled or not for each agent based on property, default_params_dict
  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    lane_pos = config_param_object["LanePosition", "Which lane should agent be at", 0]
    where_on_lane = config_param_object["WhereOnLane", "Where on lane with placeholder num_agents", "math.floor(num_agents/2)"]
    lane_positions = kwargs["agent_lane_positions"]
    lane_pos_agent_idx = [idx for idx, e in enumerate(lane_positions) if e == lane_pos]
    controlled_list_idx = eval(where_on_lane.replace("num_agents", "{}".format(len(lane_pos_agent_idx))))
    controlled_list = [False] * len(agent_states)
    controlled_list[controlled_list_idx] = True
    return controlled_list, {}, config_param_object

class AgentIds(ConfigReaderControlledAgents):
  # returns list of size num agents with true or false depending if agent is controlled or not for each agent based on property, default_params_dict
  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    controlled_ids = config_param_object["ControlledIds", "Agent ids which should be controlled", [0, 2]]
    agent_ids = kwargs["agent_ids"]
    controlled_list = []
    for id in agent_ids:
      if id in controlled_ids: 
        controlled_list.append(True)
      else:
        controlled_list.append(False)
    return controlled_list, {}, config_param_object



# Copyright (c) 2020 Patrick Hart, Julian Bernhard, 
# Klemens Esterle, Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import gym
from gym import Space
import numpy as np


class Discrete(gym.spaces.Discrete):
    def __init__(self, n):
        self._n = n
        super(Discrete, self).__init__( n)

    def sample(self):
        return self.np_random.randint(self._n)

    @property
    def low(self):
        return 0 

    @property
    def high(self):
        return self._n - 1

    def contains(self, x):
        if isinstance(x, int):
            as_int = x
        elif isinstance(x, (np.generic, np.ndarray)) and \
         (x.dtype.kind in np.typecodes['AllInteger'] and x.shape == ()):
            as_int = int(x)
        else:
            return False
        return as_int >= 0 and as_int < self._n

    def __repr__(self):
        return "Discrete(%d)" % self._n

    def __eq__(self, other):
        return isinstance(other, Discrete) and self._n == other._n


class BoundedContinuous(Space):
  def __init__(self,
               n,
               low=None,
               high=None):
    self._n = n
    self._low = low
    self._high = high
    Space.__init__(self, shape=(n,))

  def sample(self):
    if len(self._low) > 1 and len(self._high) > 1:
      sample_vec = []
      for mi, ma in zip(self._low, self._high):
        sample_vec.append(self.np_random.uniform(mi, ma))
      return np.hstack(sample_vec)
    return self.np_random.uniform(size=(self._n,))

  @property
  def low(self):
    return self._low

  @property
  def high(self):
    return self._high

  @property
  def n(self):
    return self._n

  def __repr__(self):
      return "BoundedContinuous(%d)" % self._n

  def __eq__(self, other):
    return isinstance(other, BoundedContinuous) and self._n == other._n


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Marco Oliva
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import gin
import tensorflow as tf  # pylint: disable=g-explicit-tensorflow-version-import

from tf_agents.networks import network, utils
from bark.runtime.commons.parameters import ParameterServer
from bark_ml.library_wrappers.lib_tf_agents.networks.gnns.interaction_wrapper import InteractionWrapper


@gin.configurable
class GNNCriticNetwork(network.Network):
  """Creates a critic network."""

  def __init__(self,
               input_tensor_spec,
               gnn,
               observation_fc_layer_params=None,
               observation_dropout_layer_params=None,
               observation_conv_layer_params=None,
               observation_activation_fn=tf.nn.relu,
               action_fc_layer_params=None,
               action_dropout_layer_params=None,
               action_conv_layer_params=None,
               action_activation_fn=tf.nn.relu,
               joint_fc_layer_params=None,
               joint_dropout_layer_params=None,
               joint_activation_fn=tf.nn.relu,
               output_activation_fn=None,
               name='CriticNetwork',
               params=ParameterServer()):
    """Creates an instance of `GNNCriticNetwork`.

    Args:
      input_tensor_spec: A tuple of (observation, action) each a nest of
        `tensor_spec.TensorSpec` representing the inputs.
      gnn: The function that initializes a graph neural network that 
        accepts the input observations and computes node embeddings.
      observation_fc_layer_params: Optional list of fully connected parameters
        for observations, where each item is the number of units in the layer.
      observation_dropout_layer_params: Optional list of dropout layer
        parameters, each item is the fraction of input units to drop or a
        dictionary of parameters according to the keras.Dropout documentation.
        The additional parameter `permanent', if set to True, allows to apply
        dropout at inference for approximated Bayesian inference. The dropout
        layers are interleaved with the fully connected layers; there is a
        dropout layer after each fully connected layer, except if the entry in
        the list is None. This list must have the same length of
        observation_fc_layer_params, or be None.
      observation_conv_layer_params: Optional list of convolution layer
        parameters for observations, where each item is a length-three tuple
        indicating (embedding_size, kernel_size, stride).
      observation_activation_fn: Activation function applied to the observation 
        layers, e.g. tf.nn.relu, slim.leaky_relu, ...
      action_fc_layer_params: Optional list of fully connected parameters for
        actions, where each item is the number of units in the layer.
      action_dropout_layer_params: Optional list of dropout layer parameters,
        each item is the fraction of input units to drop or a dictionary of
        parameters according to the keras.Dropout documentation. The additional
        parameter `permanent', if set to True, allows to apply dropout at
        inference for approximated Bayesian inference. The dropout layers are
        interleaved with the fully connected layers; there is a dropout layer
        after each fully connected layer, except if the entry in the list is
        None. This list must have the same length of action_fc_layer_params, or
        be None.
      action_conv_layer_params: Optional list of convolution layer
        parameters for actions, where each item is a length-three tuple
        indicating (embedding_size, kernel_size, stride).
      action_activation_fn: Activation function applied to the action layers,
        e.g. tf.nn.relu, slim.leaky_relu, ...
      joint_fc_layer_params: Optional list of fully connected parameters after
        merging observations and actions, where each item is the number of units
        in the layer.
      joint_dropout_layer_params: Optional list of dropout layer parameters,
        each item is the fraction of input units to drop or a dictionary of
        parameters according to the keras.Dropout documentation. The additional
        parameter `permanent', if set to True, allows to apply dropout at
        inference for approximated Bayesian inference. The dropout layers are
        interleaved with the fully connected layers; there is a dropout layer
        after each fully connected layer, except if the entry in the list is
        None. This list must have the same length of joint_fc_layer_params, or
        be None.
      joint_activation_fn: Activation function applied to the joint layers,
        e.g. tf.nn.relu, slim.leaky_relu, ...
      output_activation_fn: Activation function for the last layer. This can be
        used to restrict the range of the output. For example, one can pass
        tf.keras.activations.sigmoid here to restrict the output to be bounded
        between 0 and 1.
      name: A string representing name of the network.

    Raises:
      ValueError: If `observation_spec` or `action_spec` contains more than one
        observation.
    """
    super(GNNCriticNetwork, self).__init__(
      input_tensor_spec=input_tensor_spec,
      state_spec=(),
      name=name)

    observation_spec, action_spec = input_tensor_spec

    if len(tf.nest.flatten(observation_spec)) > 1:
      raise ValueError('Only a single observation is supported by this network')

    if len(tf.nest.flatten(action_spec)) > 1:
      raise ValueError('Only a single action is supported by this network')
    
    if gnn is None:
      raise ValueError('`gnn` must not be `None`.')

    self._gnn = gnn(name=name, params=params)
    
    self._observation_layers = utils.mlp_layers(
      observation_conv_layer_params,
      observation_fc_layer_params,
      observation_dropout_layer_params,
      activation_fn=observation_activation_fn,
      kernel_initializer=tf.compat.v1.keras.initializers.VarianceScaling(
        scale=1./3., mode='fan_in', distribution='uniform'),
      name='observation_encoding')

    self._action_layers = utils.mlp_layers(
      action_conv_layer_params,
      action_fc_layer_params,
      action_dropout_layer_params,
      activation_fn=action_activation_fn,
      kernel_initializer=tf.compat.v1.keras.initializers.VarianceScaling(
        scale=1./3., mode='fan_in', distribution='uniform'),
      name='action_encoding')

    self._joint_layers = utils.mlp_layers(
        None,
        joint_fc_layer_params,
        joint_dropout_layer_params,
        activation_fn=joint_activation_fn,
        kernel_initializer=tf.compat.v1.keras.initializers.VarianceScaling(
          scale=1./3., mode='fan_in', distribution='uniform'),
        name='joint_mlp')

    self._joint_layers.append(
      tf.keras.layers.Dense(
        units=1,
        activation=output_activation_fn,
        kernel_initializer=tf.keras.initializers.RandomUniform(
          minval=-0.003, maxval=0.003),
        name='value'))

  def call(self, inputs, step_type=(), network_state=(), training=False):
    del step_type # unused.

    observations, actions = inputs
    batch_size = tf.shape(observations)[0]
    embeddings = self._gnn(observations, training=training)
    
    if batch_size > 0:
      embeddings = embeddings[:, 0] # extract ego state
      actions = tf.reshape(actions, [batch_size, -1])

    # with tf.name_scope("GNNCriticEmbeddings"):
    #   tf.summary.histogram("critic_gnn_output", embeddings)
      
    embeddings = tf.cast(tf.nest.flatten(embeddings)[0], tf.float32)
    for layer in self._observation_layers:
      embeddings = layer(embeddings, training=training)

    actions = tf.cast(tf.nest.flatten(actions)[0], tf.float32)
    for layer in self._action_layers:
      actions = layer(actions, training=training)

    joint = tf.concat([embeddings, actions], 1)
    for layer in self._joint_layers:
      joint = layer(joint, training=training)

    # with tf.name_scope("GNNCriticNetwork"):
    #   tf.summary.histogram("critic_output_value", joint)

    return tf.reshape(joint, [-1]), network_state



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from gym import spaces
import numpy as np
from bark.core.models.dynamic import StateDefinition
from bark.core.world import World, ObservedWorld
from bark.runtime.commons.parameters import ParameterServer
import math
import operator

from bark_ml.observers.observer import StateObserver


class NearestAgentsObserver(StateObserver):
  def __init__(self, params=ParameterServer()):
    StateObserver.__init__(self, params)
    self._state_definition = [int(StateDefinition.X_POSITION),
                              int(StateDefinition.Y_POSITION),
                              int(StateDefinition.THETA_POSITION),
                              int(StateDefinition.VEL_POSITION)]
    self._max_distance_other_agents = \
      self._params["ML"]["NearestAgentsObserver"]["MaxOtherDistance",
      "Agents further than this distance are not observed; if not max" + \
      "other agents are seen, remaining concatenation state is set to zero",
      100]

  def Observe(self, observed_world):
    """see base class
    """
    ego_observed_world = observed_world
    num_other_agents = len(ego_observed_world.other_agents)
    ego_state = ego_observed_world.ego_agent.state

    # calculate nearest agent distances
    nearest_distances = {}
    for agent_id, agent in ego_observed_world.other_agents.items():
      if agent_id == observed_world.ego_agent.id:
        continue
      dx = ego_state[int(StateDefinition.X_POSITION)] - \
        agent.state[int(StateDefinition.X_POSITION)]
      dy = ego_state[int(StateDefinition.Y_POSITION)] - \
        agent.state[int(StateDefinition.Y_POSITION)]
      dist =  dx**2 + dy**2
      nearest_distances[dist] = agent_id

    # preallocate np.array and add ego state
    concatenated_state = np.zeros(self._len_ego_state + \
      self._max_num_vehicles*self._len_relative_agent_state, dtype=np.float32)
    concatenated_state[0:self._len_ego_state] = \
      self._select_state_by_index(self._norm(ego_state)) 
    
    # add max number of agents to state concatenation vector
    concat_pos = self._len_relative_agent_state
    nearest_distances = sorted(nearest_distances.items(),
                               key=operator.itemgetter(0))
    for agent_idx in range(0, self._max_num_vehicles):
      if agent_idx<len(nearest_distances) and \
        nearest_distances[agent_idx][0] <= self._max_distance_other_agents**2:
        agent_id = nearest_distances[agent_idx][1]
        agent = ego_observed_world.other_agents[agent_id]
        agent_rel_state = self._select_state_by_index(
          self._calculate_relative_agent_state(ego_state,
                                               self._norm(agent.state)))
        concatenated_state[concat_pos:concat_pos + \
          self._len_relative_agent_state] = agent_rel_state
      else:
        concatenated_state[concat_pos:concat_pos + \
          self._len_relative_agent_state] = \
            np.zeros(self._len_relative_agent_state)
      concat_pos += self._len_relative_agent_state
    return concatenated_state

  @property
  def observation_space(self):
    # TODO(@hart): use from spaces.py
    return spaces.Box(
      low=np.zeros(self._len_ego_state + \
        self._max_num_vehicles*self._len_relative_agent_state),
      high = np.ones(self._len_ego_state + \
        self._max_num_vehicles*self._len_relative_agent_state))

  def _norm(self, agent_state):
    if not self._normalization_enabled:
        return agent_state
    agent_state[int(StateDefinition.X_POSITION)] = \
      self._norm_to_range(agent_state[int(StateDefinition.X_POSITION)],
                          self._world_x_range)
    agent_state[int(StateDefinition.Y_POSITION)] = \
      self._norm_to_range(agent_state[int(StateDefinition.Y_POSITION)],
                          self._world_y_range)
    agent_state[int(StateDefinition.THETA_POSITION)] = \
      self._norm_to_range(agent_state[int(StateDefinition.THETA_POSITION)],
                          self._theta_range)
    agent_state[int(StateDefinition.VEL_POSITION)] = \
      self._norm_to_range(agent_state[int(StateDefinition.VEL_POSITION)],
                          self._velocity_range)
    return agent_state

  def _norm_to_range(self, value, range):
    return (value - range[0])/(range[1]-range[0])

  def _calculate_relative_agent_state(self, ego_agent_state, agent_state):
    return agent_state

  @property
  def _len_relative_agent_state(self):
    return len(self._state_definition)

  @property
  def _len_ego_state(self):
    return len(self._state_definition)


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from gym import spaces
import numpy as np
import math
import operator

from bark.core.models.dynamic import StateDefinition
from bark.core.world import ObservedWorld
from bark.runtime.commons.parameters import ParameterServer
from bark_ml.observers.observer import StateObserver


class SimpleObserver(StateObserver):
  def __init__(self,
               normalize_observations=True,
               params=ParameterServer()):
    StateObserver.__init__(self, params)
    self._state_definition = [int(StateDefinition.X_POSITION),
                              int(StateDefinition.Y_POSITION),
                              int(StateDefinition.THETA_POSITION),
                              int(StateDefinition.VEL_POSITION)]
    self._observation_len = \
      self._max_num_vehicles*self._len_state
    self._normalize_observations = normalize_observations

  def Observe(self, world):
    """see base class
    """
    concatenated_state = np.zeros(self._observation_len, dtype=np.float32)
    for i, (_, agent) in enumerate(world.agents.items()):
      state = agent.state
      if self._normalize_observations:
        state = self._normalize(state)
      reduced_state = self._select_state_by_index(state)
      # use x,y,vx,vy
      new_state = np.copy(reduced_state)
      new_state[2] = reduced_state[3]*np.cos(reduced_state[2])
      new_state[3] = reduced_state[3]*np.sin(reduced_state[2])
      starts_id = i*self._len_state
      concatenated_state[starts_id:starts_id+self._len_state] = new_state
      if i >= self._max_num_vehicles:
        break
    return concatenated_state

  def _norm(self, agent_state, position, range):
    agent_state[int(position)] = \
      (agent_state[int(position)] - range[0])/(range[1]-range[0])
    return agent_state

  def _normalize(self, agent_state):
    agent_state = \
      self._norm(agent_state,
                 StateDefinition.X_POSITION,
                 self._world_x_range)
    agent_state = \
      self._norm(agent_state,
                 StateDefinition.Y_POSITION,
                 self._world_y_range)
    agent_state = \
      self._norm(agent_state,
                 StateDefinition.THETA_POSITION,
                 self._theta_range)
    agent_state = \
      self._norm(agent_state,
                 StateDefinition.VEL_POSITION,
                 self._velocity_range)
    return agent_state

  def Reset(self, world, controlled_agent_ids):
    return world

  @property
  def observation_space(self):
    return spaces.Box(
      low=np.zeros(self._observation_len),
      high=np.ones(self._observation_len))

  @property
  def _len_state(self):
    return len(self._state_definition)





# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Marco Oliva
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import logging
import tensorflow as tf
from enum import Enum

from bark.runtime.commons.parameters import ParameterServer
from bark_ml.observers.graph_observer import GraphObserver


class GraphNetwork(tf.keras.Model):
  """
  Implements a graph neural network.

  This class serves as a wrapper over the specific implementation
  of the configured GNN library.
  """

  def __init__(self,
               params=ParameterServer(),
               name='GNN',
               output_dtype=tf.float32):
    """
    Initializes a GraphNetwork instance.

    Args:
    params: A `ParameterServer` instance containing the parameters
      to configure the GNN.
    graph_dims: A tuple containing the three elements
      (num_nodes, len_node_features, len_edge_features) of the input graph.
      Needed to properly convert observations back into a graph structure 
      that can be processed by the GNN.
    name: Name of the instance.
    output_dtype: The dtype to which the GNN output is casted.
    """
    super(GraphNetwork, self).__init__(name=name)
    self.output_dtype = output_dtype
    self._params = params
    try:
      self._graph_dims = self._validated_graph_dims(params["ML"]["GraphDims"])
    except:
      pass
    
    # logging.info(
    #   f'"{name}" configured with `spektral` for input graphs with ' +
    #   f'{self._graph_dims[0]} nodes, {self._graph_dims[1]} node features, ' + 
    #   f'and {self._graph_dims[2]} edge features.')


  def _validated_graph_dims(self, graph_dims):
    if graph_dims is None:
      raise ValueError('Graph dimensions must not be `None`.')
    if len(graph_dims) != 3:
      raise ValueError('Graph dimensions must be of length 3.')
    int_dims = list(map(int, graph_dims))
    if min(int_dims) < 0:
      raise ValueError('Graph dimensions must be positive.')
    return int_dims

  def _init_network(self):
    pass
  
  # @tf.function
  def _init_call_func(self, observations, training=False):
    pass
  
  # relax shapes due to varying batch sizes
  # @tf.function(experimental_relax_shapes=True)
  def call(self, observations, training=False):
    if tf.shape(observations)[0] == 0:
      return tf.random.normal(shape=(0, self._embedding_size))
    output = self._call_func(observations, training=training)
    return tf.cast(output, self.output_dtype)



#!/usr/bin/env python

try:
    import lanelet2
    use_lanelet2_lib = True
except:
    import warnings
    string = "Could not import lanelet2. It must be built and sourced, " + \
             "see https://github.com/fzi-forschungszentrum-informatik/Lanelet2 for details."
    warnings.warn(string)
    print("Using visualization without lanelet2.")
    use_lanelet2_lib = False
    from utils import map_vis_without_lanelet

import argparse
import os
import time
import matplotlib.pyplot as plt
from matplotlib.widgets import Button

from utils import dataset_reader
from utils import dataset_types
from utils import map_vis_lanelet2
from utils import tracks_vis
from utils import dict_utils


def update_plot():
    global fig, timestamp, title_text, track_dictionary, patches_dict, text_dict, axes
    # update text and tracks based on current timestamp
    assert(timestamp <= timestamp_max), "timestamp=%i" % timestamp
    assert(timestamp >= timestamp_min), "timestamp=%i" % timestamp
    assert(timestamp % dataset_types.DELTA_TIMESTAMP_MS == 0), "timestamp=%i" % timestamp
    title_text.set_text("\nts = {}".format(timestamp))
    tracks_vis.update_objects_plot(timestamp, track_dictionary, patches_dict, text_dict, axes)
    fig.canvas.draw()


def start_playback():
    global timestamp, timestamp_min, timestamp_max, playback_stopped
    playback_stopped = False
    plt.ion()
    while timestamp < timestamp_max and not playback_stopped:
        timestamp += dataset_types.DELTA_TIMESTAMP_MS
        start_time = time.time()
        update_plot()
        end_time = time.time()
        diff_time = end_time - start_time
        plt.pause(max(0.001, dataset_types.DELTA_TIMESTAMP_MS / 1000. - diff_time))
    plt.ioff()


class FrameControlButton(object):
    def __init__(self, position, label):
        self.ax = plt.axes(position)
        self.label = label
        self.button = Button(self.ax, label)
        self.button.on_clicked(self.on_click)

    def on_click(self, event):
        global timestamp, timestamp_min, timestamp_max, playback_stopped

        if self.label == "play":
            if not playback_stopped:
                return
            else:
                start_playback()
                return
        playback_stopped = True
        if self.label == "<<":
            timestamp -= 10*dataset_types.DELTA_TIMESTAMP_MS
        elif self.label == "<":
            timestamp -= dataset_types.DELTA_TIMESTAMP_MS
        elif self.label == ">":
            timestamp += dataset_types.DELTA_TIMESTAMP_MS
        elif self.label == ">>":
            timestamp += 10*dataset_types.DELTA_TIMESTAMP_MS
        timestamp = min(timestamp, timestamp_max)
        timestamp = max(timestamp, timestamp_min)
        update_plot()


if __name__ == "__main__":

    # provide data to be visualized
    parser = argparse.ArgumentParser()
    parser.add_argument("scenario_name", type=str, help="Name of the scenario (to identify map and folder for track "
                        "files)", nargs="?")
    parser.add_argument("track_file_number", type=int, help="Number of the track file (int)", default=0, nargs="?")
    parser.add_argument("--start_timestamp", type=int, nargs="?")
    args = parser.parse_args()

    if args.scenario_name is None:
        raise IOError("You must specify a scenario. Type --help for help.")

    # check folders and files
    error_string = ""
    tracks_dir = "../recorded_trackfiles"
    maps_dir = "../maps"
    lanelet_map_ending = ".osm"
    lanelet_map_file = maps_dir + "/" + args.scenario_name + lanelet_map_ending
    scenario_dir = tracks_dir + "/" + args.scenario_name
    track_file_prefix = "vehicle_tracks_"
    track_file_ending = ".csv"
    track_file_name = scenario_dir + "/" + track_file_prefix + str(args.track_file_number).zfill(3) + track_file_ending
    if not os.path.isdir(tracks_dir):
        error_string += "Did not find track file directory \"" + tracks_dir + "\"\n"
    if not os.path.isdir(maps_dir):
        error_string += "Did not find map file directory \"" + tracks_dir + "\"\n"
    if not os.path.isdir(scenario_dir):
        error_string += "Did not find scenario directory \"" + scenario_dir + "\"\n"
    if not os.path.isfile(lanelet_map_file):
        error_string += "Did not find lanelet map file \"" + lanelet_map_file + "\"\n"
    if not os.path.isfile(track_file_name):
        error_string += "Did not find track file \"" + track_file_name + "\"\n"
    if error_string != "":
        error_string += "Type --help for help."
        raise IOError(error_string)

    # create a figure
    fig, axes = plt.subplots(1, 1)
    fig.canvas.set_window_title("Interaction Dataset Visualization")

    # load and draw the lanelet2 map, either with or without the lanelet2 library
    lat_origin = 0.  # origin is necessary to correctly project the lat lon values in the osm file to the local
    lon_origin = 0.  # coordinates in which the tracks are provided; we decided to use (0|0) for every scenario
    print("Loading map...")
    if use_lanelet2_lib:
        projector = lanelet2.projection.UtmProjector(lanelet2.io.Origin(lat_origin, lon_origin))
        laneletmap = lanelet2.io.load(lanelet_map_file, projector)
        map_vis_lanelet2.draw_lanelet_map(laneletmap, axes)
    else:
        map_vis_without_lanelet.draw_map_without_lanelet(lanelet_map_file, axes, lat_origin, lon_origin)

    # load the tracks
    print("Loading tracks...")
    track_dictionary = dataset_reader.read_tracks(track_file_name)

    timestamp_min = 1e9
    timestamp_max = 0
    for key, track in dict_utils.get_item_iterator(track_dictionary):
        timestamp_min = min(timestamp_min, track.time_stamp_ms_first)
        timestamp_max = max(timestamp_max, track.time_stamp_ms_last)

    if args.start_timestamp is None:
        args.start_timestamp = timestamp_min

    button_pp = FrameControlButton([0.2, 0.05, 0.05, 0.05], '<<')
    button_p = FrameControlButton([0.27, 0.05, 0.05, 0.05], '<')
    button_f = FrameControlButton([0.4, 0.05, 0.05, 0.05], '>')
    button_ff = FrameControlButton([0.47, 0.05, 0.05, 0.05], '>>')

    button_play = FrameControlButton([0.6, 0.05, 0.1, 0.05], 'play')
    button_pause = FrameControlButton([0.71, 0.05, 0.1, 0.05], 'pause')

    # storage for track visualization
    patches_dict = dict()
    text_dict = dict()

    # visualize tracks
    print("Plotting...")
    timestamp = args.start_timestamp
    title_text = fig.suptitle("")
    playback_stopped = True
    update_plot()
    plt.show()



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import unittest
import numpy as np
import os
import matplotlib
import time
import tensorflow as tf

# Bark imports
from bark.runtime.commons.parameters import ParameterServer

# Bark-ml imports
from bark_ml.environments.blueprints import ContinuousHighwayBlueprint, DiscreteHighwayBlueprint
from bark_ml.environments.single_agent_runtime import SingleAgentRuntime
from bark_ml.observers.nearest_state_observer import NearestAgentsObserver
from bark_ml.observers.graph_observer_v2 import GraphObserverV2
from bark_ml.observers.graph_observer import GraphObserver
from bark_ml.core.observers import NearestObserver

from graph_nets import utils_np


class PyObserverTests(unittest.TestCase):

  def test_nearest_observer(self):
    params = ParameterServer()
    bp = ContinuousHighwayBlueprint(params)
    env = SingleAgentRuntime(blueprint=bp, render=True)
    env.reset()
    world = env._world

    # under test
    observer = NearestAgentsObserver(params)

    eval_id = env._scenario._eval_agent_ids[0]
    observed_world = world.Observe([eval_id])[0]
    start_time = time.time()
    observed_state = observer.Observe(observed_world)
    end_time = time.time()
    print(f"It took {end_time-start_time} seconds.")
    print(observed_state, observer.observation_space.shape)
    
  def test_nearest_observer_cpp(self):
    params = ParameterServer()
    bp = ContinuousHighwayBlueprint(params)
    env = SingleAgentRuntime(blueprint=bp, render=True)
    env.reset()
    world = env._world

    # under test
    observer = NearestObserver(params)
    observer.Reset(world)
    
    eval_id = env._scenario._eval_agent_ids[0]
    observed_world = world.Observe([eval_id])[0]
    start_time = time.time()
    observed_state = observer.Observe(observed_world)
    end_time = time.time()
    print(f"It took {end_time-start_time} seconds.")
    print(observed_state, observer.observation_space.shape)


  def test_graph_observer_v2(self):
    params = ParameterServer()
    bp = ContinuousHighwayBlueprint(params)
    env = SingleAgentRuntime(blueprint=bp, render=True)
    env.reset()
    world = env._world
    params["ML"]["GraphObserver"]["AgentLimit"] = 5
    
    # observers
    observer = GraphObserver(params)
    reference_observer = GraphObserverV2(params)
    reference_observer.Reset(world)
    observer.Reset(world)
    
    
    eval_id = env._scenario._eval_agent_ids[0]
    observed_world = world.Observe([eval_id])[0]
    start_time = time.time()
    observed_state = observer.Observe(observed_world)
    observed_ref_state = reference_observer.Observe(observed_world)
    end_time = time.time()
    print(f"It took {end_time-start_time} seconds.")

    # batch
    batch_observation = tf.stack([observed_state, observed_state], axis=0)
    graph = GraphObserver.graph(
      batch_observation, graph_dims=observer.graph_dimensions, dense=True)
    
    batch_observation_reference = tf.stack(
      [observed_ref_state, observed_ref_state], axis=0)
    graph_ref = GraphObserverV2.graph(
      batch_observation_reference, dense=True)
    
    # unpack values
    ref_node_vals, ref_edge_indices, ref_node_lens, ref_edge_lens, ref_globals, ref_edge_vals = graph_ref
    node_vals, edge_indices, _, edge_vals = graph
    
    # TODO: assert this
    print(ref_node_vals, node_vals)
    # print(ref_edge_vals, edge_vals)
    print(ref_edge_indices, ref_edge_vals)
    print(edge_indices, edge_vals)


if __name__ == '__main__':
  unittest.main()


from .matplotlib_viewer import MPViewer
from .pygame_viewer import PygameViewer
from .video_renderer import VideoRenderer
from .buffered_viewer import BufferedViewer
from .viewer import BaseViewer

__all__ = ["MPViewer","PygameViewer","VideoRenderer","BufferedViewer","BaseViewer"]



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.
import sys
import logging
import time
import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt
tf.compat.v1.enable_v2_behavior()

# BARK imports
from bark.runtime.commons.parameters import ParameterServer
from bark.runtime.viewer.buffered_viewer import BufferedViewer
from bark.core.geometry import *
from bark.core.world.renderer import *

# tf agent imports
from tf_agents.drivers import dynamic_step_driver
from tf_agents.drivers import dynamic_episode_driver
from tf_agents.environments import tf_py_environment
from tf_agents.metrics import tf_metrics
from tf_agents.eval import metric_utils
from tf_agents.utils import common
from tf_agents.trajectories import time_step as ts
from tf_agents.agents.sac import sac_agent

# BARK-ML imports
from bark_ml.library_wrappers.lib_tf_agents.py_bark_environment import PyBARKEnvironment
from bark_ml.commons.tracer import Tracer


class TFARunner:
  def __init__(self,
               environment=None,
               agent=None,
               tracer=None,
               params=None):
    self._params = params or ParameterServer()
    self._eval_metrics = [
      tf_metrics.AverageReturnMetric(
        buffer_size=self._params["ML"]["TFARunner"]["EvaluationSteps", "", 25]),
      tf_metrics.AverageEpisodeLengthMetric(
        buffer_size=self._params["ML"]["TFARunner"]["EvaluationSteps", "", 25])
    ]
    self._agent = agent
    self._agent.set_action_externally = True
    self._summary_writer = None
    self._environment = environment
    self._wrapped_env = tf_py_environment.TFPyEnvironment(
      PyBARKEnvironment(self._environment))
    self.GetInitialCollectionDriver()
    self.GetCollectionDriver()
    self._logger = logging.getLogger()
    self._tracer = tracer or Tracer()
    self._colliding_scenario_ids = []
    
  def SetupSummaryWriter(self):
    if self._params["ML"]["TFARunner"]["SummaryPath"] is not None:
      try:
        self._summary_writer = tf.summary.create_file_writer(
          self._params["ML"]["TFARunner"]["SummaryPath"])
      except:
        pass
    self.GetInitialCollectionDriver()
    self.GetCollectionDriver()

  def GetInitialCollectionDriver(self):
    self._initial_collection_driver = \
      dynamic_episode_driver.DynamicEpisodeDriver(
        env=self._wrapped_env,
        policy=self._agent._agent.collect_policy,
        observers=[self._agent._replay_buffer.add_batch],
        num_episodes=self._params["ML"]["TFARunner"]["InitialCollectionEpisodes", "", 50])

  def GetCollectionDriver(self):
    self._collection_driver = dynamic_episode_driver.DynamicEpisodeDriver(
      env=self._wrapped_env,
      policy=self._agent._agent.collect_policy,
      observers=[self._agent._replay_buffer.add_batch],
      num_episodes=self._params["ML"]["TFARunner"]["CollectionEpisodesPerStep", "", 1])

  def CollectInitialEpisodes(self):
    self._initial_collection_driver.run()

  def Train(self):
    self.CollectInitialEpisodes()
    if self._summary_writer is not None:
      with self._summary_writer.as_default():
        self._train()
    else:
      self._train()

  def _train(self):
    """Agent specific
    """
    pass

  def ReshapeActionIfRequired(self, action_step):
    action_shape = action_step.action.shape
    expected_shape = self._agent._eval_policy.action_spec.shape
    action = action_step.action.numpy()
    if action_shape != expected_shape:
      # logging.warning("Action shape" + str(action_shape) + \
      #   " does not match with expected shape " + str(expected_shape) +\
      #   " -> reshaping is tried")
      action = np.reshape(action, expected_shape)
      # logging.info(action)
    return action
  
  @staticmethod
  def _id_to_idx(id_agent_id_map, aid):
    return list(id_agent_id_map.keys())[list(id_agent_id_map.values()).index(aid)]
  
  @staticmethod
  def _id_agent_map(world, obs, ego_id):
    ego_agent = world.agents[ego_id]
    agents = list(world.agents.values())
    agents.remove(ego_agent)
    agents = obs._agents_sorted_by_distance(ego_agent, agents)
    agents.insert(0, ego_agent)
    agents = agents[:obs._num_agents]
    id_agent_id_map = {}
    for idx, agent in enumerate(agents):
      id_agent_id_map[idx] = agent.id
    return id_agent_id_map

  @staticmethod
  def _get_agent_pos(world, aid):
    agent = world.agents[aid]
    return [agent.state[1], agent.state[2]]
    
  def ProcessGraphTuple(self, env, graph_tuple, ego_id, render=False):
    senders = graph_tuple.senders.numpy()
    receivers = graph_tuple.receivers.numpy()
    edges = graph_tuple.edges.numpy()
    
    # sorted list
    id_agent_id_map = self._id_agent_map(env._world, env._observer, ego_id)
    # get idx
    for idx, agent_id in id_agent_id_map.items():
      receiver_idx = np.where(
        receivers == self._id_to_idx(id_agent_id_map, ego_id))
      sender_idx = senders[receiver_idx]
      for sid in sender_idx:
        from_id = agent_id
        to_id = id_agent_id_map[sid]
        from_pos = self._get_agent_pos(env._world, from_id)
        to_pos = self._get_agent_pos(env._world, to_id)
        # TODO: use sum(..), sum(abs(..)), min(..), max(..), ...
        norm = np.sum(np.fabs(edges), axis=(1, 0))
        magnitude = np.sum(edges[sid])/norm
        color = (128/255, 128/255, 128/255, .5)
        zorder = 1
        alpha = 0.4
        if agent_id == ego_id:
          color = (12/255, 44/255, 132/255, 1.)
          zorder = 10
          alpha = 0.8
        length = np.sqrt((from_pos[0] - to_pos[0])**2 + (from_pos[1] - to_pos[1])**2)
        from_agent_state = env._world.agents[from_id].state
        to_agent_state = env._world.agents[to_id].state
        if to_id != agent_id:
          self._tracer.Trace(
            {"length_dx": (from_pos[0] - to_pos[0]),
            "length_dy": (from_pos[1] - to_pos[1]),
            "dv": from_agent_state[4] - to_agent_state[4],
            # "dtheta": SignedAngleDiff(from_agent_state[3], to_agent_state[3]),
            "magnitude": np.sum(edges[sid])})
        if render:
          if isinstance(env._viewer, BufferedViewer):
            l = Line2d()
            l.AddPoint(Point2d(from_pos[0], from_pos[1]))
            l.AddPoint(Point2d(to_pos[0], to_pos[1]))
            line_primitive = RenderPrimitive(l)
            if agent_id == ego_id:
              line_primitive.Add("stroke_color", [12, 44, 132, 255])
            else:
              line_primitive.Add("stroke_color", [128, 128, 128, 128])
            line_primitive.Add("stroke_width", max(5*magnitude, .1))
            env._world.renderer.Add("LINES", line_primitive)
          else:
            ax = env._viewer.axes      
            ax.plot(
              [from_pos[0], to_pos[0]],
              [from_pos[1], to_pos[1]],
              color=color, marker="o", linewidth=max(min(50*magnitude, 5.), 0.5), zorder=zorder, alpha=alpha)
    
    if render:
      for idx, agent_id in id_agent_id_map.items():
        agent_pos = self._get_agent_pos(env._world, agent_id)
        color = "gray"
        if agent_id == ego_id:
          color = "blue"
        if isinstance(env._viewer, BufferedViewer):
          p = Point2d(agent_pos[0], agent_pos[1])
          point_primitive = RenderPrimitive(p)
          point_primitive.Add("stroke_color", [12, 44, 132, 255])
          point_primitive.Add("radius_pixels", 0.3)
          point_primitive.Add("fill_color", [12, 44, 132, 128])
          env._world.renderer.Add("POINTS", point_primitive)
        else:
          ax = env._viewer.axes
          ax.plot(
            agent_pos[0], agent_pos[1], marker='o', color=color, markersize=6)
      
  def RunEpisode(self, render=True, trace_colliding_ids=None, **kwargs):
    state = self._environment.reset()
    is_terminal = False
    # print(self._agent._eval_policy.trainable_variables)
    if render:
      self._environment.render()
    while not is_terminal:
      action_step = self._agent._eval_policy.action(
        ts.transition(state, reward=0.0, discount=1.0))
      action = self.ReshapeActionIfRequired(action_step)
      env_data = self._environment.step(action)
      self._tracer.Trace(env_data, **kwargs)
      state, reward, is_terminal, info = env_data
      # graph stuff
      try:
        graph_tuples = self._agent._agent._actor_network._latent_trace
        ego_id = self._environment._scenario._eval_agent_ids[0]
        self.ProcessGraphTuple(
          self._environment,
          graph_tuples[-1],
          ego_id,
          render)
      except:
        pass
      if render:
        self._logger.info(f"The ego agent's action is {action} and " + \
                          f"a reward of {reward}.")
        self._environment.render()
      if is_terminal and (info["collision"] or info["drivable_area"]) and trace_colliding_ids is not None:
        self._colliding_scenario_ids.append(
          self._environment._scenario_idx)
      if is_terminal and info["goal_reached"]:
        self._logger.info("\033[92mThe ego agent reached its goal. \033[0m")
      
  def Run(
    self, num_episodes=10, render=False, mode="not_training",
    trace_colliding_ids=None, **kwargs):
    for i in range(0, num_episodes):
      if render:
        self._logger.info(f"Simulating episode {i}.")
      trajectory = self.RunEpisode(
        render=render, num_episode=i, trace_colliding_ids=trace_colliding_ids, **kwargs)
    mean_col_rate = self._tracer.collision_rate
    goal_reached = self._tracer.success_rate
    mean_reward = self._tracer.mean_reward
    mean_steps = self._tracer.mean_steps
  
    if mode == "training":
      global_iteration = self._agent._agent._train_step_counter.numpy()
      tf.summary.scalar("mean_reward", mean_reward, step=global_iteration)
      tf.summary.scalar("mean_steps", mean_steps, step=global_iteration)
      tf.summary.scalar(
        "mean_collision_rate", mean_col_rate, step=global_iteration)
      tf.summary.scalar(
        "goal_reached", goal_reached, step=global_iteration)
    print(
      f"The agent achieved an average reward of {mean_reward:.3f}," +
      f" collision-rate of {mean_col_rate:.5f}, took on average" +
      f" {mean_steps:.3f} steps, and reached the goal " + 
      f" {goal_reached:.3f} (evaluated over {num_episodes} episodes).")
    if trace_colliding_ids:
      return self._colliding_scenario_ids
    else:
      return {
        "goal_reached": goal_reached,
        "mean_col_rate": mean_col_rate,
        "mean_steps": mean_steps,
        "mean_reward": mean_reward,
      }


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import os
from bark.runtime.commons.parameters import ParameterServer
from bark.runtime.viewer.matplotlib_viewer import MPViewer
from bark.runtime.scenario.scenario_generation.config_with_ease import \
  LaneCorridorConfig, ConfigWithEase
from bark.core.models.dynamic import SingleTrackModel
from bark.core.world.opendrive import XodrDrivingDirection
from bark.core.world.goal_definition import GoalDefinitionStateLimitsFrenet

from bark_ml.environments.blueprints.blueprint import Blueprint
from bark_ml.evaluators.goal_reached import GoalReached
from bark_ml.evaluators.goal_reached_guiding import GoalReachedGuiding
from bark_ml.behaviors.cont_behavior import BehaviorContinuousML
from bark_ml.behaviors.discrete_behavior import BehaviorDiscreteMotionPrimitivesML, \
        BehaviorDiscreteMacroActionsML

from bark_ml.behaviors.discrete_behavior import BehaviorDiscreteMotionPrimitivesML, \
        BehaviorDiscreteMacroActionsML
from bark_ml.core.observers import NearestObserver


class HighwayLaneCorridorConfig(LaneCorridorConfig):
  def __init__(self,
               params=None,
               **kwargs):
    super(HighwayLaneCorridorConfig, self).__init__(params, **kwargs)
  
  def goal(self, world):
    road_corr = world.map.GetRoadCorridor(
      self._road_ids, XodrDrivingDirection.forward)
    lane_corr = self._road_corridor.lane_corridors[0]
    return GoalDefinitionStateLimitsFrenet(lane_corr.center_line,
                                           (0.4, 0.4),
                                           (0.1, 0.1),
                                           (12.5, 17.5))


class HighwayBlueprint(Blueprint):
  def __init__(self,
               params=None,
               num_scenarios=250,
               random_seed=0,
               ml_behavior=None,
               viewer=True,
               mode="dense"):
    if mode == "dense":
      ds_min = 10.
      ds_max = 20.
    if mode == "medium":
      ds_min = 20.
      ds_max = 35.
    params["BehaviorIDMClassic"]["DesiredVelocity"] = 15.
    params["World"]["remove_agents_out_of_map"] = False
    left_lane = HighwayLaneCorridorConfig(params=params,
                                          road_ids=[16],
                                          lane_corridor_id=0,
                                          min_vel=12.5,
                                          max_vel=17.5,
                                          ds_min=ds_min,
                                          ds_max=ds_max,
                                          s_min=5.,
                                          s_max=200.,
                                          controlled_ids=None)
    right_lane = HighwayLaneCorridorConfig(params=params,
                                           road_ids=[16],
                                           lane_corridor_id=1,
                                           min_vel=12.5,
                                           max_vel=17.5,
                                           s_min=5.,
                                           s_max=200.,
                                           ds_min=ds_min,
                                           ds_max=ds_max,
                                           controlled_ids=True)
    scenario_generation = \
      ConfigWithEase(
        num_scenarios=num_scenarios,
        map_file_name=os.path.join(os.path.dirname(__file__), "../../../environments/blueprints/highway/city_highway_straight.xodr"),  # NOLINT
        random_seed=random_seed,
        params=params,
        lane_corridor_configs=[left_lane, right_lane])
    if viewer:
      viewer = MPViewer(params=params,
                        x_range=[-50, 50],
                        y_range=[-50, 50],
                        follow_agent_id=True)
    dt = 0.2
    evaluator = GoalReached(params)
    observer = NearestObserver(params)
    ml_behavior = ml_behavior

    super().__init__(
      scenario_generation=scenario_generation,
      viewer=viewer,
      dt=dt,
      evaluator=evaluator,
      observer=observer,
      ml_behavior=ml_behavior)


class ContinuousHighwayBlueprint(HighwayBlueprint):
  def __init__(self,
               params=None,
               num_scenarios=25,
               random_seed=0,
               viewer=True,
               mode="dense"):
    ml_behavior = BehaviorContinuousML(params)
    HighwayBlueprint.__init__(self,
                              params=params,
                              num_scenarios=num_scenarios,
                              random_seed=random_seed,
                              ml_behavior=ml_behavior,
                              viewer=True,
                              mode=mode)


class DiscreteHighwayBlueprint(HighwayBlueprint):
  def __init__(self,
               params=None,
               num_scenarios=25,
               random_seed=0,
               mode="dense"):
    ml_behavior = BehaviorDiscreteMacroActionsML(params)
    HighwayBlueprint.__init__(self,
                              params=params,
                              num_scenarios=num_scenarios,
                              random_seed=random_seed,
                              ml_behavior=ml_behavior,
                              viewer=True,
                              mode=mode)


from bark.core.geometry import *

def GetGoalLaneCorridorForGoal(observed_world):
  """Returns the lanecorridor the goal is in"""
  ego_agent = observed_world.ego_agent
  goal_def = ego_agent.goal_definition
  goal_shape = goal_def.goal_shape
  rc = observed_world.ego_agent.road_corridor
  lane_corr = None
  for lc in rc.lane_corridors:
    if Collide(lc.polygon, goal_shape):
      lane_corr = lc
  return lane_corr

def CalculateDistanceToGoal(observed_world):
  """Calculates the distance to the goal of the ego_agent"""
  goal_lane_corr = GetGoalLaneCorridorForGoal(observed_world)
  goal_center_line = goal_lane_corr.center_line
  ego_agent = observed_world.ego_agent
  ego_agent_state = ego_agent.state
  distance_to_goal = Distance(
    goal_center_line,
    Point2d(ego_agent_state[1], ego_agent_state[2]))
  return distance_to_goal


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from abc import ABC, abstractmethod
from bark.runtime.commons.parameters import ParameterServer


class StateEvaluator(ABC):
  """Evaluates the state of the environment
     e.g., if a collision has happend
  """
  def __init__(self,
               params=ParameterServer()):
    self._params = params
    self._viewer = None

  def Evaluate(self, observed_world, action):
    """Evaluates the observed world
    """
    eval_results, reward, done = None, 0., False
    eval_results = observed_world.Evaluate()
    reward, done, eval_results = self._evaluate(
      observed_world, eval_results, action)
    return reward, done, eval_results

  def Reset(self, world):
    world.ClearEvaluators()
    evaluators = self._add_evaluators()
    for key, evaluator in evaluators.items():
      world.AddEvaluator(key, evaluator)
    return world

  def SetViewer(self, viewer):
    self._viewer = viewer


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Marco Oliva, Silvan Wimmer
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import numpy as np
import logging
import tensorflow as tf
from gym import spaces
from typing import Dict

from bark.core.models.dynamic import StateDefinition
from bark.core.geometry import Distance, Point2d
from bark.runtime.commons.parameters import ParameterServer

from bark_ml.observers.observer import StateObserver

class GraphObserver(StateObserver):
  """
  This observer converts an `ObservsedWorld` instance into
  a graph-structured observation, consisting of nodes (with
  features), an adjacency matrix expressing the (directed)
  connections between nodes in the graph, and edges (with
  features), expressing the relative information between
  the agents that constitute these connections.
  """

  def __init__(self, params=ParameterServer()):
    """
    Creates an instance of `GraphObserver`.

    Args:
      params: A `ParameterServer` instance that enables further
        configuration of the observer. Defaults to a new instance.
    """
    StateObserver.__init__(self, params)
    self._logger = logging.getLogger()

    self._normalize_observations = \
      self._params["ML"]["GraphObserver"]["NormalizationEnabled",
      "Whether normalization of features should be performed.",
      True]

    self._num_agents =\
      params["ML"]["GraphObserver"]["AgentLimit", 
      "The maximum number of agents that is observed.",
      4]
    self._visibility_radius =\
      params["ML"]["GraphObserver"]["VisibilityRadius",
      "The radius in which agent can 'see', i.e. detect other agents.", 
      550]
    self._add_self_loops =\
      params["ML"]["GraphObserver"]["SelfLoops", 
      "Whether nodes have self connections.", 
      True]
    requested_node_attribute_keys =\
      params["ML"]["GraphObserver"]["EnabledNodeFeatures",
      "The list of available node features, given by their string key that \
       the observer should extract from the world and insert into the \
       observation. For a list of available features, refer to the list \
       returned by `GraphObserver.available_node_attributes`.",
      self.available_node_attributes()
    ]
    self._enabled_node_attribute_keys = self._filter_requested_attributes(
      requested_keys=requested_node_attribute_keys,
      available_keys=self.available_node_attributes(),
      context="node")
    self._logger.info(
      "GraphObserver configured with node attributes: " +
      f"{self._enabled_node_attribute_keys}")
    requested_edge_attribute_keys =\
      params["ML"]["GraphObserver"]["EnabledEdgeFeatures",
      "The list of available edge features, given by their string key that \
       the observer should extract from the world and insert into the \
       observation. For a list of available features, refer to the list \
       returned by `GraphObserver.available_edge_attributes`.",
      self.available_edge_attributes()
    ]
    self._enabled_edge_attribute_keys = self._filter_requested_attributes(
      requested_keys=requested_edge_attribute_keys,
      available_keys=self.available_edge_attributes(),
      context="edge")
    self._logger.info(
      "GraphObserver configured with edge attributes: " +
      f"{self._enabled_edge_attribute_keys}")

    # the number of features of a node in the graph
    self.feature_len = len(self._enabled_node_attribute_keys)
    # the number of features of an edge between two nodes
    self.edge_feature_len = len(self._enabled_edge_attribute_keys)

  def Observe(self, world):
    """see base class"""
    agents = self._preprocess_agents(world)

    # placeholder for the output observation
    obs = np.zeros(self._len_state)

    # insert node features for each agent
    for i, agent in agents:
      start_index = i * self.feature_len
      end_index = start_index + self.feature_len
      obs[start_index:end_index] = self._extract_node_features(agent)

    edge_features = np.zeros((self._num_agents, self._num_agents, self.edge_feature_len))
    adjacency_matrix = np.zeros((self._num_agents, self._num_agents))

    # add edges to all visible agents
    for index, agent in agents:
      nearby_agents = self._nearby_agents(
        center_agent=agent, 
        agents=agents, 
        radius=self._visibility_radius)

      if self._add_self_loops:
        adjacency_matrix[index, index] = 1

      for target_index, nearby_agent in nearby_agents:
        edge_features[index, target_index, :] = self._extract_edge_features(agent, nearby_agent)
        adjacency_matrix[index, target_index] = 1
        adjacency_matrix[target_index, index] = 1
    
    # insert adjacency list
    adj_start_index = self._num_agents * self.feature_len
    adj_end_index = adj_start_index + self._num_agents ** 2
    obs[adj_start_index:adj_end_index] = adjacency_matrix.reshape(-1)

    # insert edge features
    obs[adj_end_index:] = edge_features.reshape(-1)
    
    return tf.convert_to_tensor(obs, dtype=tf.float32, name='observation')

  @classmethod
  def graph(cls, observations, graph_dims, dense=False):
    """
    Maps the given batch of observations into a
    graph representation.

    Args:
    observations: The batch of observations as a tf.Tensor
      of shape (batch_size, observation_size).
    graph_dims: A tuple containing the dimensions of the 
      graph as (num_nodes, num_features, num_edge_features).
      If `dense` is set to True, num_edge_features is ignored.
    dense: Specifies the format of the returned graph representation. 
      If set to `True`, the edges are returned as a list of pairs 
      of nodes indices (relative to the flattened batch) and an additional
      mapping of each node to a graph is returned. If set to 
      False (default), edges are returned as sparse adjacency 
      matrix and an edge feature matrix is additionally returned.

    Returns:
      X: Node features of (batch_size, num_nodes, num_features)

      If `dense` is True:
      A: Dense representation of edges as a list of node index pairs,
        shape (num_total_edges, 2).
      node_to_graph_map: A 1d list, where each element is the mapping
      of the node (indexed relative to the batch) to a graph.
      
      If `dense`is False:
      A: Spare binary adjacency matrix of shape 
        (batch_size num_nodes, num_nodes).
      E: Edge features of shape 
        (batch_size, num_nodes, num_edge_features, num_edge_features).
    """
    obs = observations # for brevity

    if not tf.is_tensor(obs):
      obs = tf.convert_to_tensor(obs)
    
    n_nodes, n_features = graph_dims[0:2]
    batch_size = tf.shape(observations)[0]
    
    # extract node features F
    F = tf.reshape(obs[:, :n_nodes * n_features], [batch_size, n_nodes, n_features])

    # extract adjacency matrix A
    adj_start_idx = n_nodes * n_features
    adj_end_idx = adj_start_idx + n_nodes ** 2
    A = tf.reshape(obs[:, adj_start_idx:adj_end_idx], [batch_size, n_nodes, n_nodes])

  
    if dense:
      # in dense mode, the nodes of all graphs are 
      # concatenated into one list that contains all 
      # feature vectors of the whole batch.
      # the assignment to graphs is handled by the
      # node_to_graph_map.
      F = tf.reshape(F, [batch_size * n_nodes, n_features])

      # find non-zero elements in the adjacency matrix (edges)
      # and collect their indices
      A = tf.where(tf.greater(A, 0))

      # we need the indices of the source and target nodes to
      # be represented as their indices in the whole batch,
      # in other words: each node index must be the index
      # of the graph in the batch plus the index of the node
      # in the graph. E.g. if each graph has 5 nodes, the 
      # node indices are: graph 0: 0-4, graph 1: 5-9, etc.

      # compute a tensor where each element
      # is the graph index of the node that is represented
      # by the same index in A
      graph_indices = tf.reshape(A[:, 0], [1, -1])
      graph_indices = tf.scalar_mul(n_nodes, graph_indices)
      graph_indices = tf.transpose(tf.tile(graph_indices, [2, 1]))
      
      A = A[:, 1:] + graph_indices
      A = tf.cast(A, tf.int32)

      # since the node feature vectors are all stored in a 2D array,
      # we need to map them to the graphs that they belong to.

      # construct a mapping from the node feature vectors in F
      # to the graphs in the batch, i.e a list where each element 
      # contains the index of the graph that the feature vector at
      # the same index in F belongs to, e.g. if each graph contains
      # two nodes, then the node_to_graph_map is [0, 0, 1, 1, 2, 2, ...]
      node_to_graph_map = tf.reshape(tf.range(batch_size), [-1, 1])
      node_to_graph_map = tf.tile(node_to_graph_map, [1, n_nodes])
      node_to_graph_map = tf.reshape(node_to_graph_map, [-1])
      
      # extract edge features
      n_edge_features = graph_dims[2]
      E_shape = [-1, n_edge_features]
      E = tf.reshape(obs[:, adj_end_idx:], E_shape)
      return F, A, node_to_graph_map, E

    # extract edge features
    n_edge_features = graph_dims[2]
    E_shape = [batch_size, n_nodes, n_nodes, n_edge_features]
    E = tf.reshape(obs[:, adj_end_idx:], E_shape)
    return F, A, E

  def _preprocess_agents(self, world):
    """
    Preproccesses the agents for constructing an
    observation by resorting.

    Args:
      world: A `bark.core.ObservedWorld` instance.

    Returns:
      A list of tuples, consisting of an index and an 
      agent object element.

      The first element always represents the ego agent.
      The remaining elements resemble other agents, up 
      to the limit defined by `self._num_agents`, sorted 
      in ascending order with respect to the agents'
      distance in the world to the ego agent.
    """
    ego_agent = world.ego_agent
    agents = list(world.agents.values())
    agents.remove(ego_agent)
    agents = self._agents_sorted_by_distance(ego_agent, agents)
    agents.insert(0, ego_agent)
    return list(enumerate(agents))[:self._num_agents]

  def _agents_sorted_by_distance(self, ego_agent, agents):
    """
    Returns the given list of `agents`, sorted in ascending
    order by their relative distance to the `ego_agent`.
    """
    def distance(agent):
      return Distance(
        self._position(ego_agent), 
        self._position(agent))
    if len(agents) > 2:
      agents.sort(key=distance)
    return agents

  def _nearby_agents(self, center_agent, agents, radius: float):
    """
    Returns all elements from 'agents' whose position, defined 
    in x and y coordinates is within the specified `radius` of 
    the 'center_agent's position.
    """
    center_agent_pos = self._position(center_agent)
    other_agents = filter(lambda a: a[1].id != center_agent.id, agents)
    nearby_agents = []
    for index, agent in other_agents:
      agent_pos = self._position(agent)
      distance = Distance(center_agent_pos, agent_pos)
      if distance <= radius:
        nearby_agents.append((index, agent))
    return nearby_agents

  def _extract_node_features(self, agent, as_dict=False):
    res = {}

    # Init data (to keep ordering always equal for reading and writing)
    for label in self._enabled_node_attribute_keys:
      res[label] = "inf"
    
    n = self.normalization_data
    def add_feature(key, value, norm_range):
      if self._normalize_observations:
        value = self._normalize_value(value, norm_range)
      res[key] = value

    try:
      state = agent.state
      if "x" in res:
        add_feature("x", state[int(StateDefinition.X_POSITION)], n["x"])
      if "y" in res:
        add_feature("y", state[int(StateDefinition.Y_POSITION)], n["y"])
      if "theta" in res:
        add_feature("theta", state[int(StateDefinition.THETA_POSITION)], n["theta"])
      if "vel" in res:
        add_feature("vel", state[int(StateDefinition.VEL_POSITION)], n["vel"])

      # get information related to goal
      goal_center = agent.goal_definition.goal_shape.center[0:2]
      goal_dx = goal_center[0] - state[int(StateDefinition.X_POSITION)]
      goal_dy = goal_center[1] - state[int(StateDefinition.Y_POSITION)]

      if "goal_x" in res:
        add_feature("goal_x", goal_center[0], n["x"])
      if "goal_y" in res:
        add_feature("goal_y", goal_center[1], n["y"])
      if "goal_dx" in res:
        add_feature("goal_dx", goal_dx, n["dx"])
      if "goal_dy" in res:
        add_feature("goal_dy", goal_dy, n["dy"])
      if "goal_theta" in res:
        add_feature("goal_theta", np.arctan2(goal_dy, goal_dx), n["theta"])
      if "goal_d" in res:
        add_feature("goal_d", np.sqrt(goal_dx**2 + goal_dy**2), n["distance"])
      if "goal_vel" in res:
        goal_velocity = np.mean(agent.goal_definition.velocity_range)
        add_feature("goal_vel", goal_velocity, n["vel"])
    except:
      raise AttributeError(
        "A problem occured during node feature extraction. Possibly " +
        "a node feature that's specifed in the 'EnabledNodeFeatures' " +
        "parameter is not supported by the current BARK-ML environment.")
    
    # remove disabled attributes
    res = {key: res[key] for key in self._enabled_node_attribute_keys}
    assert list(res.keys()) == self._enabled_node_attribute_keys

    if not as_dict:
      res = list(res.values())

    return res

  def _extract_edge_features(self, source_agent, target_agent):
    """
    Encodes the relation between the given agents as the
    features of a directed edge in the graph.

    Args:
      source_agent: the agent that defines the source of the edge.
      target_agent: the agent that defines the target of the edge.

    Returns:
      An np.array containing the differences in the agents' 
      x and y position, velocities and orientations.
    """
    source_features =\
      self._extract_node_features(source_agent, as_dict=True)
    target_features =\
      self._extract_node_features(target_agent, as_dict=True)

    features = {
      "dx": source_features["x"] - target_features["x"],
      "dy": source_features["y"] - target_features["y"],
      "dvel": source_features["vel"] - target_features["vel"],
      "dtheta": source_features["theta"] - target_features["theta"]
    }

    features = {key: features[key] for key in self._enabled_edge_attribute_keys}
    features = list(features.values())
    assert len(features) == len(self._enabled_edge_attribute_keys)

    return features

  def _filter_requested_attributes(self, 
                                   requested_keys, 
                                   available_keys, 
                                   context=""):
    """
    Filters the requested node/edge attributes by the available 
    attributes and returns the intersection of the requested
    and available attributes.
    """
    if not isinstance(requested_keys, list):
      raise ValueError(
        f"Requested {context} feature list must be a list of " +
        f"strings, not {type(requested_keys)}.")

    if len(requested_keys) == 0:
      self._logger.warning(
        "The `GraphObserver` received an empty list of requested " +
        f"{context} attributes.")
      return []

    valid_keys = []
    invalid_keys = []

    for key in requested_keys:
      if key in available_keys:
        valid_keys.append(key)
      else:
        invalid_keys.append(key)
    
    if len(invalid_keys) > 0:
      self._logger.warning(
        f"The following {context} attributes requested from the GraphObserver " +
        f"are not supported and will be ignored: {invalid_keys}")

    return valid_keys

  def _normalize_value(self, value, range):
    """
    Normalizes the `value` with can take on values in
    into the range of [-1, 1].
    If the `value` is outside the given range, it's clamped 
    to the bound of [-1, 1]
    """
    normed = (value - range[0]) / (range[1] - range[0])
    # normed = max(-1, normed) # values lower -1 clipped
    # normed = min(1, normed) # values bigger 1 clipped
    return normed

  def reset(self, world):
    return world

  def _position(self, agent) -> Point2d:
    return Point2d(
      agent.state[int(StateDefinition.X_POSITION)],
      agent.state[int(StateDefinition.Y_POSITION)]
    )

  @property
  def normalization_data(self) -> Dict[str, list]:
    """
    The reference ranges of how certain attributes are normalized.
    Use this info to scale normalized values back to real values.
    
    E.g. the value for key 'x' returns the possible range of 
    the x-element of positions. A normalized value of 1.0 
    corresponds to the maximum of this range (and vice versa).

    Note: This dictionary does not include a range for each
    possible attribute, but rather for each _kind_ of attribute,
    like distances, velocities, angles, etc.
    E.g. all distances (between agents, between objects, etc.)
    are scaled relative to the 'distance' range. 
    """ 
    x_range = self._world_x_range[1] - self._world_x_range[0]
    y_range = self._world_y_range[1] - self._world_y_range[0]
    max_dist = np.linalg.norm([x_range, y_range])

    d = {}
    d['x'] = self._world_x_range
    d['y'] = self._world_y_range
    d['theta'] = self._theta_range
    d['vel'] = self._velocity_range
    d['distance'] = [0, max_dist]
    d['dx'] = [-x_range, x_range]
    d['dy'] = [-y_range, y_range]
    return d

  def sample(self):
    raise NotImplementedError

  @classmethod
  def available_node_attributes(cls, with_descriptions=False):
    attributes = {
      "x": "The x-components of the agent's position.",
      "y": "The y-components of the agent's position.", 
      "theta": "The current heading angle of tha agent.",
      "vel": "The current velocity of the agent.", 
      "goal_x": "The x-component of the goal's position.", 
      "goal_y": "The y-component of the goal's position.",
      "goal_dx": "The difference in the x-component of the agent's and the goal's position.", 
      "goal_dy": "The difference in the y-component of the agent's and the goal's position.", 
      "goal_theta": "The goal heading angle.", 
      "goal_d": "The euclidian distance of the agent to the goal.", 
      "goal_vel": "The goal velocity."
    }

    if not with_descriptions:
      attributes = list(attributes.keys())

    return attributes

  @classmethod
  def available_edge_attributes(cls, with_descriptions=False):
    attributes = {
      "dx": "The difference in the x-position of the two agents.", 
      "dy": "The difference in the y-position of the two agents.", 
      "dvel": "The difference in the velocity of the two agents.", 
      "dtheta": "The difference in the heading angle of the two agents."
    }

    if not with_descriptions:
      attributes = list(attributes.keys())

    return attributes

  @property
  def observation_space(self):    
    # 0 ... 1   for all node attributes
    # 0 ... 1   for the adjacency list
    # 0 ... 1   for the edge attributes
    return spaces.Box(
      low=np.concatenate((
        np.zeros(self._num_agents*self.feature_len),
        np.zeros(self._num_agents**2),
        np.zeros(self._num_agents**2*self.edge_feature_len))),
      high=np.ones(self._len_state))

  @property
  def _len_state(self):
    len_node_features = self._num_agents*self.feature_len
    len_adjacency = self._num_agents**2
    len_edge_features = len_adjacency*self.edge_feature_len
    return len_node_features + len_adjacency + len_edge_features

  @property
  def graph_dimensions(self):
    """
    Returns a three-element tuple that defines the
    dimensions of the graph as returned by the
    `Observe` method, consisting of
    - the number of nodes
    - the number of node features
    - the number of edge features
    """
    return (
      self._num_agents,
      self.feature_len,
      self.edge_feature_len
    )


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import numpy as np
import logging
from bark.core.viewer import Viewer
from bark.core.geometry import *
from bark.core.models.dynamic import *
from bark.core.world.opendrive import *
from bark.core.world.goal_definition import *
from bark.runtime.commons.parameters import ParameterServer
import math
from bark.core.world.evaluation.ltl import *
from bark.core.world.evaluation import *
from bark.core.models.behavior import *

logger = logging.getLogger()


class BaseViewer(Viewer):
    def __init__(self, params=None, **kwargs):
        if(params is None):
            params = ParameterServer()
        Viewer.__init__(self)
        # color parameters
        # agents
        self.color_other_agents_line = params["Visualization"]["Agents"]["Color"]["Other"]["Lines",
                                                                                           "Color of other agents", (0.1, 0.1, 0.1)]
        self.color_other_agents_face = params["Visualization"]["Agents"]["Color"]["Other"]["Face",
                                                                                           "Color of other agents", (0.7, 0.7, 0.7)]
        self.color_eval_agents_line = params["Visualization"]["Agents"]["Color"]["Controlled"]["Lines",
                                                                                               "Color of controlled, evaluated agents", (0.9, 0, 0)]
        self.color_eval_agents_face = params["Visualization"]["Agents"]["Color"]["Controlled"]["Face",
                                                                                               "Color of controlled, evaluated agents", (0.9, 0, 0)]
        self.use_colormap_for_other_agents = params["Visualization"]["Agents"]["Color"][
            "UseColormapForOtherAgents", "Flag to enable color map for other agents", True]
        self.if_colormap_use_line_others = params["Visualization"]["Agents"]["Color"]["IfColormapUseLineColorOthers",
                                                                                      "Flag to enable that line color can be fixed for other agents while using colormap", True]
        self.alpha_eval_agent = params["Visualization"]["Agents"]["Alpha"]["Controlled",
                                                                           "Alpha of evalagents", 0.8]
        self.alpha_other_agents = params["Visualization"]["Agents"]["Alpha"]["Other",
                                                                             "Alpha of other agents", 1]
        self.route_color = params["Visualization"]["Agents"]["ColorRoute",
                                                             "Color of agents routes", (0.2, 0.2, 0.2)]
        self.draw_route = params["Visualization"]["Agents"]["DrawRoute",
                                                            "Draw Route of each agent", False]
        self.draw_agent_id = params["Visualization"]["Agents"]["DrawAgentId",
                                                               "Draw id of each agent", True]
        self.draw_invalid_agents = params["Visualization"]["Agents"]["DrawInvalidAgents", "Draw invalid agents with patch", False]
        self.draw_orientation_arrow = params["Visualization"]["Agents"]["DrawOrientationArrow",
                                                               "Draw Orientation of Arrow", False]
        self.draw_behavior_plan_eval_agent = params["Visualization"]["Agents"]["DrawBehaviorPlanEvalAgent", "Draw behavior plan of evalauted agent", False]
        self.draw_eval_goals = params["Visualization"]["Agents"]["DrawEvalGoals",
                                                                 "Draw Route of eval agent goals", True]
        self.eval_goal_color = params["Visualization"]["Agents"]["EvalGoalColor",
                                                                 "Color of eval agent goals", (.49, .63, .83)]
        self.draw_history = params["Visualization"]["Agents"]["DrawHistory",
                                                              "Draw history with alpha trace for each agent", False]
        self.draw_history_draw_face = params["Visualization"]["Agents"]["DrawHistoryDrawFace",
                                                                        "Flag to specify if face is drawn in history mode", True]

        # map
        self.color_lane_boundaries = params["Visualization"]["Map"]["XodrLanes"]["Boundaries"]["Color",
                                                                                               "Color of agents except ego vehicle", (0.7, 0.7, 0.7)]
        self.alpha_lane_boundaries = params["Visualization"]["Map"]["XodrLanes"][
            "Boundaries"]["Alpha", "Color of agents except ego vehicle", 1.0]
        self.plane_color = params["Visualization"]["Map"]["Plane"]["Color",
                                                                   "Color of the background plane", (1, 1, 1, 1)]
        self.plane_alpha = params["Visualization"]["Map"]["Plane"]["Alpha",
                                                                   "Alpha of the background plane", 1.0]
        self.map_linewidth = params["Visualization"]["Map"]["XodrLanes"]["Boundaries"]["Linewidth",
                                                                                       "Linewidth of linestrings", 1.0]
        self._draw_aerial_image = params["Visualization"]["Map"]["DrawAerialImage", "Flag to draw aerial image behind map", False]

        self.draw_ltl_debug_info = params["Visualization"]["Evaluation"]["DrawLTLDebugInfo",
                                                                         "Flag to specify if debug info to ltl evaluators shall be plotted", False]
                                                                        
        self.draw_rss_debug_info = params["Visualization"]["Evaluation"]["DrawRssDebugInfo",
                                                                         "Flag to specify if debug info to rss evaluators shall be plotted", False]

        self.draw_rss_safety_responses = params["Visualization"]["Evaluation"]["DrawRssSafetyResponses",
                                                                               "Flag to specify if visualizating rss safety responses.", False]

        self._draw_ego_rss_safety_responses = params["Visualization"]["Evaluation"][
          "DrawEgoRSSSafetyResponses",
          "Flag to specify if visualizating rss safety responses.",
          False]
        self._rss_min_braking_distances = params["Visualization"]["Evaluation"][
          "DrawMinRSSBrakingDistances",
          "Flag whether the min. braking distances shall be plotted.",
          False]
        
        self.parameters = params
        self.agent_color_map = {}
        self.max_agents_color_map = 0

        self.use_world_bounds = kwargs.pop("use_world_bounds", False)
        self.follow_agent_id = kwargs.pop("follow_agent_id", None)

        self.center = kwargs.pop("center", np.array([0, 0]))

        self.world_x_range = kwargs.pop("x_range", np.array([-40, 40]))
        self.world_y_range = kwargs.pop("y_range", np.array([-40, 40]))

        self.enforce_x_length = kwargs.pop("enforce_x_length", True)
        self.enforce_y_length = kwargs.pop("enforce_y_length", False)
        self.x_length = kwargs.pop(
            "x_length", np.sum(np.absolute(self.world_x_range)))
        self.y_length = kwargs.pop(
            "y_length", np.sum(np.absolute(self.world_y_range)))

        self.dynamic_world_x_range = self.world_x_range.copy()
        self.dynamic_world_y_range = self.world_y_range.copy()

    def reset():
        pass

    def get_aspect_ratio(self):
        pass

    def _get_draw_eval_agent_ids(self, world, eval_agent_ids=None, ):
        if self.follow_agent_id is not None:
            if isinstance(self.follow_agent_id, bool) and \
                    eval_agent_ids is not None and \
                    len(eval_agent_ids) == 1:
                draw_eval_agent_id = eval_agent_ids[0]
            else:
                draw_eval_agent_id = self.follow_agent_id

            if draw_eval_agent_id in world.agents:
                return draw_eval_agent_id

        return None

    def _update_world_view_range(self, world, eval_agent_ids=None):
        draw_eval_agent_id = self._get_draw_eval_agent_ids(
            world, eval_agent_ids)

        if draw_eval_agent_id != None:
            follow_agent = world.agents[draw_eval_agent_id]
            state = follow_agent.state
            pose = generatePoseFromState(state)

            center = [pose[0],  pose[1]]
            self._update_world_dynamic_range(center)
        else:
            if self.use_world_bounds:
                bb = world.bounding_box
                self.dynamic_world_x_range = [bb[0].x(), bb[1].x()]
                self.dynamic_world_y_range = [bb[0].y(), bb[1].y()]

                diffx = abs(
                    self.dynamic_world_x_range[1] - self.dynamic_world_x_range[0])
                diffy = abs(
                    self.dynamic_world_y_range[1] - self.dynamic_world_y_range[0])

                # enforce that in both dimensions  the same range is covered
                if diffx > diffy:
                    self.dynamic_world_y_range[0] -= (diffx - diffy)/2
                    self.dynamic_world_y_range[1] += (diffx - diffy)/2
                else:
                    self.dynamic_world_x_range[0] -= (diffy - diffx)/2
                    self.dynamic_world_x_range[1] += (diffy - diffx)/2
            else:
                center = self.center
                self._update_world_dynamic_range(center)

    def _update_world_dynamic_range(self, center):
        aspect_ratio = self.get_aspect_ratio()
        if self.enforce_x_length:
            self.dynamic_world_x_range = [-self.x_length /
                                          2 + center[0], self.x_length/2 + center[0]]
            self.dynamic_world_y_range = [-self.x_length/2/aspect_ratio +
                                          center[1], self.x_length/2/aspect_ratio + center[1]]

        if self.enforce_y_length:
            self.dynamic_world_x_range = [-self.y_length/2*aspect_ratio +
                                          center[0], self.y_length/2*aspect_ratio + center[0]]
            self.dynamic_world_y_range = [-self.y_length /
                                          2 + center[1], self.y_length/2 + center[1]]

    def drawPoint2d(self, point2d, color, alpha):
        pass

    def drawLine2d(self, line2d, color, alpha, line_style=None, zorder=2):
        pass

    def drawPolygon2d(self, polygon, color, alpha, facecolor=None, zorder=10, hatch=''):
        pass

    def drawTrajectory(self, trajectory, color, **kwargs):
        pass
    
    def drawArrow(self, pose):
        pass

    def drawObstacle(self, obstacle):
        pass

    def drawText(self, position, text, **kwargs):
        pass

    def drawCircles(self, position_list, radius):
        for pos in position_list:
          self.drawCircle(pos, radius)

    def drawCircle(self, position, radius):
        pass

    def getColor(self, color):
        pass

    def show(self, block=False):
        pass

    def clear(self):
        pass

    def drawAgents(self, world):
        for _, agent in world.agents.items():
            if agent.id in world.agents_valid:
              self.drawAgent(agent)
            else:
              if self.draw_invalid_agents:
                self.drawAgent(agent, hatch='o')
    
    def drawBehaviorPlan(self, agent):
        self.drawTrajectory(agent.behavior_model.last_trajectory,
                                  color='black', linewidth=1.0)

    def drawHistory(self, agent, color, alpha, facecolor, zorder):
        shape = agent.shape
        if isinstance(shape, Polygon2d):
            history = agent.history
            lh = len(history)
            for idx, state_action in enumerate(history):
                state = state_action[0]
                pose = generatePoseFromState(state)
                transformed_polygon = shape.Transform(pose)
                alpha = 1-0.8*(lh-idx)/3.4
                alpha = 0 if alpha < 0 else alpha
                # fade to 0.2 after 10 steps
                self.drawPolygon2d(transformed_polygon, color,
                                   alpha, facecolor, zorder=zorder)

    def drawGoalDefinition(self, goal_definition, color, alpha, facecolor):
        if isinstance(goal_definition, GoalDefinitionPolygon):
            self.drawPolygon2d(goal_definition.goal_shape,
                               color, alpha, facecolor)
        elif isinstance(goal_definition, GoalDefinitionStateLimits):
            self.drawPolygon2d(goal_definition.xy_limits,
                               color, alpha, facecolor)
        elif isinstance(goal_definition, GoalDefinitionStateLimitsFrenet):
            self.drawPolygon2d(goal_definition.goal_shape,
                               color, alpha, facecolor)
        elif isinstance(goal_definition, GoalDefinitionSequential):
            prev_center = np.array([])
            for idx, goal_def in enumerate(goal_definition.sequential_goals):
                self.drawGoalDefinition(goal_def, color, alpha, facecolor)
                goal_pos = None
                if isinstance(goal_def, GoalDefinitionPolygon):
                    goal_pos = goal_def.goal_shape.center
                elif isinstance(goal_def, GoalDefinitionStateLimits):
                    goal_pos = goal_def.xy_limits.center
                # self.drawText(position=goal_pos, text="Goal{}".format(idx), coordinate="world")
                if prev_center.any():
                    line = Line2d()
                    line.AddPoint(Point2d(prev_center[0], prev_center[1]))
                    line.AddPoint(Point2d(goal_pos[0], goal_pos[1]))
                    self.drawLine2d(line, color, alpha=0.9)
                prev_center = goal_pos

    def drawLabelsAsText(self, observed_world, label_functions, evaluator_type):
        labels = {}
        for lf in label_functions:
            labels.update(lf.Evaluate(observed_world))
        str = ""
        for (k, v) in labels.items():
            # if k.agent_id in [-1, 104, 108]:
            if v:
                str = str + \
                    "{} {}_{} : {}\n".format(
                        evaluator_type, k.label_str, k.agent_id, v)
        self.drawText(position=(0.7, 0.9), text=str)

    def drawWorld(self, world, eval_agent_ids=None, filename=None, scenario_idx=None, debug_text=True):
        # self.clear()
        self._update_world_view_range(world, eval_agent_ids)
        if world.map:
            self.drawMap(world.map.GetOpenDriveMap())

        if self._draw_aerial_image:
          self.drawMapAerialImage()

        # draw agent goals
        for agent_id, agent in world.agents.items():
            if eval_agent_ids and self.draw_eval_goals and agent.goal_definition and \
                    agent_id in eval_agent_ids:
                color_line = self.color_eval_agents_line
                color_face = self.color_eval_agents_face
                alpha = self.alpha_eval_agent
                self.drawGoalDefinition(
                    agent.goal_definition, color_line, alpha, color_face)

        num_agents = len(world.agents.items())
        for i, (agent_id, agent) in enumerate(world.agents.items()):
            alpha = 1.0

            if eval_agent_ids and agent.id in eval_agent_ids:
                color_line = self.color_eval_agents_line
                color_face = self.color_eval_agents_face
                alpha = self.alpha_eval_agent
            else:
                alpha = self.alpha_other_agents
                if self.use_colormap_for_other_agents:
                    if not agent_id in self.agent_color_map:
                        color_idx = len(self.agent_color_map) % self.getSizeOfColormap()
                        self.agent_color_map[agent_id] = self.getColorFromMap(color_idx)
                    if self.if_colormap_use_line_others:
                        color_line = self.color_other_agents_line
                    else:
                        color_line = self.agent_color_map[agent_id]
                    color_face = self.agent_color_map[agent_id]
                else:
                    color_line = self.color_other_agents_line
                    color_face = self.color_other_agents_face
                    
            if self.draw_history and agent.id in world.agents_valid:
                if self.draw_history_draw_face:
                    color_face_history = color_face
                else:
                    color_face_history = (1.0, 1.0, 1, .0)
                self.drawHistory(agent, color_line, alpha,
                                 color_face_history, zorder=5)
            
            hatch = '' if agent.id in world.agents_valid else 'o'
            if agent.id in world.agents_valid or self.draw_invalid_agents:
              self.drawAgent(agent, color_line, alpha, color_face, hatch=hatch)
              
        if debug_text:
            self.drawText(position=(0.1, 0.9), text="Scenario: {}".format(
                scenario_idx), fontsize=14)
            self.drawText(position=(0.1, 0.95),
                          text="Time: {:.2f}".format(world.time), fontsize=14)

        if self.draw_ltl_debug_info:
            self.drawLTLDebugInfomation(world, eval_agent_ids[0])

        if self.draw_rss_debug_info or self.draw_rss_safety_responses:
            if self.draw_rss_debug_info:
                self.drawRssDebugInfomation(world, eval_agent_ids[0])
            if self.draw_rss_safety_responses:
                self.drawRssSafetyResponses(world, eval_agent_ids[0])

        if self.draw_behavior_plan_eval_agent:
          eval_agent = world.GetAgent(eval_agent_ids[0])
          if eval_agent is not None:
              self.drawBehaviorPlan(eval_agent)
        
        if self._draw_ego_rss_safety_responses:
          self.DrawRSSEvaluatorState(world, eval_agent_ids[0])
        
        if self._rss_min_braking_distances:
          self.DrawRSSBrakingDistances(world, eval_agent_ids[0])

    def drawMapAerialImage(self):
        pass

    def drawMap(self, map):
        # draw the boundary of each lane
        for _, road in map.GetRoads().items():
            self.drawXodrRoad(road, self.color_lane_boundaries)

    def drawXodrRoad(self, road, color=None):
        for lane_section in road.lane_sections:
            self.drawXodrLaneSection(lane_section, color)

    def drawXodrLaneSection(self, lane_section, color=None):
        for _, lane in lane_section.GetLanes().items():
            self.drawXodrLane(lane, color)

    def drawXodrLane(self, lane, color=None):
        if color is None:
            self.color_lane_boundaries

        dashed = False
        # center line is type none and is drawn as broken
        if lane.road_mark.type == XodrRoadMarkType.broken or lane.road_mark.type == XodrRoadMarkType.none:
            dashed = True
        self.drawLine2d(lane.line, color, self.alpha_lane_boundaries,
                        dashed, zorder=1, linewidth=self.map_linewidth)

    def drawAgent(self, agent, color, alpha, facecolor, hatch=''):
        shape = agent.shape
        if isinstance(shape, Polygon2d):
            state = agent.state
            pose = generatePoseFromState(state)
            transformed_polygon = shape.Transform(pose)

            centerx = (shape.front_dist - 0.5*(shape.front_dist +
                                               shape.rear_dist)) * math.cos(pose[2]) + pose[0]
            centery = (shape.front_dist - 0.5*(shape.front_dist +
                                               shape.rear_dist)) * math.sin(pose[2]) + pose[1]

            if self.draw_agent_id:
                self.drawText(position=(centerx, centery), rotation=180.0*(1.0+pose[2]/math.pi), text="{}".format(agent.id),
                              coordinate="not axes", ha='center', va="center", multialignment="center", size="smaller")
            
            if self.draw_orientation_arrow:
              self.drawArrow(pose)

            self.drawPolygon2d(transformed_polygon, color,
                               alpha, facecolor, zorder=10, hatch=hatch)
        else:
            raise NotImplementedError("Shape drawing not implemented.")

    def drawLaneCorridor(self, lane_corridor, color=None):
        if color is None:
            color = "blue"
        self.drawPolygon2d(lane_corridor.polygon, color=color,
                           facecolor=color, alpha=.3, zorder=2)

    def Reset(self):
        pass

    def drawRoadCorridor(self, road_corridor, color=None):
        if color is None:
            color = "blue"
        self.drawPolygon2d(road_corridor.polygon, color,
                           facecolor=color, alpha=.2, zorder=2)

    def drawLTLDebugInfomation(self, world, agent_id):
        observed_world = world.Observe([agent_id])
        if len(observed_world) == 0:
            return
        observed_world = observed_world[0]
        for _, evaluator_type in enumerate(observed_world.evaluators):
            if isinstance(observed_world.evaluators[evaluator_type], EvaluatorLTL):
                label_functions = observed_world.evaluators[evaluator_type].label_functions
                # we will only plot labels of first ltl evaluator
                self.drawLabelsAsText(
                    observed_world, label_functions, evaluator_type)
                break

    def drawRssDebugInfomation(self, world, agent_id):
        from bark.core.world.evaluation import EvaluatorRSS
        for evaluator in world.evaluators:
          if isinstance(world.evaluators[evaluator], EvaluatorRSS):
            rss_responses = world.evaluators[evaluator].PairwiseDirectionalEvaluate(
                world)
            break

        def char_func(value):
            if value == True:
                return "T"
            elif value == False:
                return "F"
            else:
                return "UNKNOWN"

        overall_safety = True
        if rss_responses:
            self.drawText(position=(0.82, 0.91), text="ID  Lon  Lat")
            for i, (id, responses) in enumerate(rss_responses.items()):
                overall_safety = overall_safety and any(responses)
                str = "{}:    {}     {}".format(
                    id, *list(map(char_func, responses)))
                self.drawText(position=(0.82, 0.88-0.03*i), text=str)

        self.drawText(position=(0.74, 0.96), horizontalalignment="left", text="ego id {} safety: {}".format(
            agent_id, char_func(overall_safety)))
    
    def drawRssSafetyResponses(self, world, ego_id):
        from bark.core.world.evaluation import EvaluatorRSS
        for evaluator in world.evaluators:
            if isinstance(world.evaluators[evaluator], EvaluatorRSS):
                rss_responses = world.evaluators[evaluator].PairwiseEvaluate(
                  world)
                break

        ego_agent = world.agents[ego_id]
        shape = ego_agent.shape
        pose = generatePoseFromState(ego_agent.state)
        transformed_polygon = shape.ScalingTransform(1.5, pose)
        self.drawPolygon2d(
            transformed_polygon,
            self.color_eval_agents_line, 0.6, self.color_other_agents_face, linewidth=1.5, zorder=9)

        # draw response for other agents
        relevant_agents = [
            agent for agent in world.agents.values() if agent.id in rss_responses]
        for agent in relevant_agents:
            shape = agent.shape
            pose = generatePoseFromState(agent.state)
            transformed_polygon = shape.ScalingTransform(1.5, pose)

            response_color = "LightGreen" if rss_responses[agent.id] else "Red"
            self.drawPolygon2d(transformed_polygon, response_color,
                               0.6, response_color, zorder=9)

    def DrawRSSBrakingDistances(self, world, eval_id):
      for agent_id, agent in world.agents.items():
        observed_world = world.Observe([agent_id])[0]
        rss_params = None
        if eval_id == agent_id:
          rss_params = self.parameters["EvaluatorRss"]["Ego"]
        else:
          rss_params = self.parameters["EvaluatorRss"]["Others"]
        try:
          min_braking_safety_polygon = ComputeMinBrakingPolygon(observed_world, rss_params)
        except:
          pass
        color_face = "red"
        self.drawPolygon2d(
          min_braking_safety_polygon.GetPolygon(), color_face, 0.25, color_face, zorder=9)
        
    def DrawRSSEvaluatorState(self, world, agent_id):
      agent = world.agents[agent_id]
      behavior = agent.behavior_model

      if isinstance(behavior, BehaviorRSSConformant):
        longitudinal_response = behavior.GetLongitudinalResponse()
        lateral_left_response = behavior.GetLateralLeftResponse()
        lateral_right_response = behavior.GetLateralRightResponse()
        violations = []
        if longitudinal_response != 0:
          print("Longitudinal violation")
          violations.append({"label": "LON", "color": "purple"})
        if lateral_left_response != 0 or lateral_right_response != 0:
          print("Lateral violation")
          violations.append({"label": "LAT", "color": "red"})
        
        # have to compute shape here
        observed_world = world.Observe([agent_id])[0]
        behavior.ComputeSafetyPolygons(observed_world)
        safety_polygons = behavior.GetSafetyPolygons()
    
        for poly in safety_polygons:
          print(poly)
          color_face = self.agent_color_map[poly.GetAgentId()]
          self.drawPolygon2d(poly.GetPolygon(), color_face, 0.3, color_face, zorder=9)
          
        # draw labels
        ego_agent = world.agents[agent_id]
        shape = ego_agent.shape
        pose = generatePoseFromState(ego_agent.state)
        a = shape.front_dist - 0.5*(shape.front_dist + shape.rear_dist)
        centerx = a * math.cos(pose[2]) + pose[0] + 2.5
        centery = a * math.sin(pose[2]) + pose[1] + shape.right_dist + 2.
        
        if self.draw_agent_id:
          for violation in violations:
            t = self.drawText(
              position=(centerx, centery), rotation=180.0*(1.0+pose[2]/math.pi),
              text="{}".format(violation["label"]),
              coordinate="not axes", ha='center', va="center",
              multialignment="center", size="smaller",
              color = violation["color"])
            centerx += 5.
            t.set_bbox(dict(
              facecolor=violation["color"], alpha=0.5,
              edgecolor=violation["color"]))


def generatePoseFromState(state):
  # pybind creates column based vectors, initialization maybe row-based -> we consider both
  pose = np.zeros(3)
  pose[0] = state[int(StateDefinition.X_POSITION)]
  pose[1] = state[int(StateDefinition.Y_POSITION)]
  pose[2] = state[int(StateDefinition.THETA_POSITION)]
  return pose



# Copyright (c) 2020 Patrick Hart, Julian Bernhard,
# Klemens Esterle, Tobias Kessler
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT
import numpy as np
from bark.core.world.evaluation import \
  EvaluatorGoalReached, EvaluatorCollisionEgoAgent, \
  EvaluatorStepCount, EvaluatorDrivableArea, CaptureAgentStates
from bark.runtime.commons.parameters import ParameterServer
from bark.core.models.dynamic import StateDefinition

from bark_ml.evaluators.evaluator import StateEvaluator
from bark_ml.evaluators.commons import *

def VelocityPotential(v, v_des, v_dev_max=10., a=0.4):
  return 1. - (np.sqrt((v-v_des)**2)/v_dev_max)**a

def DistancePotential(d, d_max, b=0.4):
  return 1. - (d/d_max)**b 

def ObjectPotential(d, d_max, c=0.4):
  return -1. + (d/d_max)**c
  
  
class RewardShapingEvaluator(StateEvaluator):
  def __init__(self,
               params=ParameterServer(),
               eval_agent=None):
    StateEvaluator.__init__(self, params)
    self._goal_reward = \
      self._params["ML"]["RewardShapingEvaluator"]["GoalReward",
        "Reward for reaching the goal.",
        1.]
    self._col_penalty = \
      self._params["ML"]["RewardShapingEvaluator"]["CollisionPenalty",
        "Reward given for a collisions.",
        -1.]
    self._max_steps = \
      self._params["ML"]["RewardShapingEvaluator"]["MaxSteps",
        "Maximum steps per episode.",
        60]
    self._active_shaping_functions = \
      self._params["ML"]["RewardShapingEvaluator"]["RewardShapingPotentials",
        "Reward shaping functions.", {
          "VelocityPotential" : {
            "desired_vel": 10., "vel_dev_max": 10., "exponent": 0.4, "type": "positive"
          },
          "DistancePotential": {
            "exponent": 0.4, "d_max": 10., "type": "positive"
          },
          "DistanceOtherPotential": {
            "exponent": 0.4, "d_max": 20., "type": "negative"
          }
        }]
    self._eval_agent = eval_agent

  def _add_evaluators(self):
    evaluators = {}
    evaluators["goal_reached"] = EvaluatorGoalReached()
    evaluators["collision"] = EvaluatorCollisionEgoAgent()
    evaluators["step_count"] = EvaluatorStepCount()
    evaluators["drivable_area"] = EvaluatorDrivableArea()
    return evaluators

  def RewardShapingFunction(self, observed_world, gamma=0.99):
    ego_agent = observed_world.ego_agent
    state_history = [state_action[0] for state_action in ego_agent.history[-2:]]
    last_state, current_state = state_history
    negative_potentials, positive_potentials = [], []
    if "VelocityPotential" in self._active_shaping_functions:
      params_vp = self._active_shaping_functions["VelocityPotential"]
      vel_potentials = [VelocityPotential(
        state[int(StateDefinition.VEL_POSITION)],
        params_vp["desired_vel"],
        v_dev_max=params_vp["vel_dev_max"],
        a=params_vp["exponent"]) for state in [last_state, current_state]]
      positive_potentials.append(vel_potentials)
    if "DistancePotential" in self._active_shaping_functions:
      params_goal = self._active_shaping_functions["DistancePotential"]
      goal_center_line =  ego_agent.goal_definition.center_line
      if Distance(
          goal_center_line,
          Point2d(current_state[1], current_state[2])) < params_goal["d_max"]:
        goal_potentials = [DistancePotential(
          Distance(
            goal_center_line,
            Point2d(state[1], state[2])),
          d_max=params_goal["d_max"],
          b=params_goal["exponent"]) for state in [last_state, current_state]]
        positive_potentials.append(goal_potentials)
    if "DistanceOtherPotential" in self._active_shaping_functions:
      params_obj = self._active_shaping_functions["DistanceOtherPotential"]
      # TODO: integrate lon lat offsets
      for agent in observed_world.other_agents.values():
        other_state_history = [state_action[0] for state_action in agent.history[-2:]]
        other_last_state, other_current_state = other_state_history
        if Distance(
          Point2d(other_current_state[1], other_current_state[2]),
          Point2d(current_state[1], current_state[2])) > params_obj["d_max"]:
          continue
        goal_potentials = [ObjectPotential(
          Distance(
            Point2d(state[0][1], state[0][2]),
            Point2d(state[1][1], state[1][2])),
          d_max=params_obj["d_max"],
          c=params_obj["exponent"]) for state in [
            (last_state, other_last_state), (current_state, other_current_state)]]
        negative_potentials.append(goal_potentials)
    
    number_pos_potentials = len(positive_potentials)
    number_neg_potentials = len(negative_potentials)
    
    # normalize
    for i in range(0, number_pos_potentials):
      positive_potentials[i][0] /= number_pos_potentials
      positive_potentials[i][1] /= number_pos_potentials
    for i in range(0, number_neg_potentials):
      negative_potentials[i][0] /= number_neg_potentials
      negative_potentials[i][1] /= number_neg_potentials
    
    # reward shaping function gamma*p_{t+1} - p_t
    reward_shaping_value = 0.
    for potential_values in positive_potentials:
      reward_shaping_value += gamma*potential_values[1] - potential_values[0]
    for potential_values in negative_potentials:
      reward_shaping_value += gamma*potential_values[1] - potential_values[0]
    return reward_shaping_value
  
  def _evaluate(self, observed_world, eval_results, action):
    """Returns information about the current world state
    """
    done = False
    success = eval_results["goal_reached"]
    step_count = eval_results["step_count"]
    collision = eval_results["collision"] or eval_results["drivable_area"]

    reward_shaping_signal = self.RewardShapingFunction(observed_world)
    if success or collision or step_count > self._max_steps:
      done = True
    
    if collision:
      success = 0
      # TODO: needs to be for logging
      eval_results["goal_reached"] = False
      
    # for now it is only collision and success
    reward = collision * self._col_penalty + \
      success * self._goal_reward + reward_shaping_signal
    return reward, done, eval_results
    
  def Reset(self, world):
    return super(RewardShapingEvaluator, self).Reset(world)


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import pickle
import os
from bark.runtime.commons.parameters import ParameterServer

class ScenarioGeneration:
  def __init__(self, params=None, num_scenarios=None, random_seed=1000):
    self._params = params
    self._current_scenario_idx = 0
    self._random_seed = random_seed

    if params is None:
        self._params = ParameterServer()
    else:
        self._params = params
    self.initialize_params(self._params)
    self._scenario_list = self.create_scenarios(params, num_scenarios)
  
  def initialize_params(self, params):
    pass

  @property
  def params(self):
      return self._params

  def get_next_scenario(self):
    if self._current_scenario_idx >= self.num_scenarios:
      self._current_scenario_idx = 0
      #print("Resetting scenario index to zero")
    scenario = self.get_scenario(self._current_scenario_idx)
    scenario_idx = self._current_scenario_idx
    self._current_scenario_idx += 1
    return scenario, scenario_idx

  def get_num_scenarios(self):
    return len(self._scenario_list)

  def get_scenario(self, idx):
    return self._scenario_list[idx].copy()

  def __iter__(self):
    self._current_iter_idx=0
    return self

  def __next__(self):
    if self._current_iter_idx < self.get_num_scenarios():
      scenario = self.get_scenario(self._current_iter_idx)
      idx = self._current_iter_idx
      self._current_iter_idx += 1
      return scenario, idx
    else:
      raise StopIteration

  def create_scenarios(self, params, num_scenarios):
    """ Creates a list of scenario class instances which should be
        deterministically reproducible given the random seed,
        the params and the number of scenarios
    
    Arguments:
        params {[bark.common.ParameterServer]} -- [provides additional parameters]
        num_scenarios {[int]} -- [how many scenarios should be created]
        random_seed {[unsigned int]} -- [seed used to make scenario generation
                                         based on random factors reproducible]
    
    Returns:
        scenario_list {[a list of instances of type scenario class]} -- 
          [each scenario in this list defines one initial world condition]
    """
    return None

  @property
  def num_scenarios(self):
    return len(self._scenario_list)

  def dump_scenario_list(self, filename):
    with open(filename, "wb") as file:
      # print("SAVE PATH:", os.path.abspath(filename))
      pickle.dump(self._scenario_list, file)

  def load_scenario_list(self, filename):
    with open(filename, "rb") as file:
      self._scenario_list = pickle.load(file)



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import pickle
from enum import Enum

class Tracer:
  def __init__(self, trace_history=True):
    self._states = []
    self._trace_history = trace_history
    self._total_episodes = 0  # increases with terminal count
    self._total_collisions = 0
    self._total_goals_reached = 0
    self._total_reward = 0
    self._total_steps = 0

  def EvalStateToDict(self, eval_state):
    """Converts an eval_state to a flat eval_dict"""
    eval_dict = {}
    if type(eval_state) is tuple:
      (state, reward, is_terminal, info) = eval_state
      eval_dict["state"] = state
      eval_dict["reward"] = reward
      eval_dict["is_terminal"] = is_terminal
      for info_key, info_value in info.items():
        eval_dict[info_key] = info_value
    if type(eval_state) is dict:
      for info_key, info_value in eval_state.items():
        eval_dict[info_key] = info_value
    return eval_dict

  def Trace(self, eval_state, **kwargs):
    """Traces and stores a state"""
    eval_dict = self.EvalStateToDict(eval_state)
    for key, value in kwargs.items():
      eval_dict[key] = value
    if self._trace_history:
      self._states.append(eval_dict)
    if "is_terminal" in eval_dict:
      if eval_dict["is_terminal"]:
        self._total_episodes += 1
    if "drivable_area" in eval_dict and "collision" in eval_dict:
      if eval_dict["drivable_area"] or eval_dict["collision"]:
        self._total_collisions += 1
    if "goal_reached" in eval_dict:
      if eval_dict["goal_reached"]:
        self._total_goals_reached += 1
    if "reward" in eval_dict:
      self._total_reward += eval_dict["reward"]
    if not "magnitude" in eval_dict:
      self._total_steps += 1
  
  @property
  def collision_rate(self):
    if self._total_episodes == 0:
      return self._total_collisions
    return self._total_collisions/self._total_episodes
  
  @property
  def success_rate(self):
    if self._total_episodes == 0:
      return self._total_goals_reached
    return self._total_goals_reached/self._total_episodes
  
  @property
  def mean_steps(self):
    if self._total_episodes == 0:
      return self._total_steps
    return self._total_steps/self._total_episodes

  @property
  def mean_reward(self):
    if self._total_episodes == 0:
      return self._total_reward
    return self._total_reward/self._total_episodes
  
  @property
  def episode_number(self):
    return self._total_episodes

  def Save(self, filepath="./"):
    """Saves trace as pandas dataframe"""
    filehandler = open(filepath, "wb")
    pickle.dump(self.__dict__, file=filehandler)
    filehandler.close()

  def Load(self, filepath="./"):
    filehandler = open(filepath, "wb")
    self.__dict__ = pickle.load(self, file=filehandler)
    filehandler.close()

  def Reset(self):
    self._states = []
    self._total_episodes = 0
    self._total_collisions = 0
    self._total_goals_reached = 0
    self._total_steps = 0
    self._total_reward = 0
    
    
    



from bark_ml.library_wrappers.lib_tf_agents.agents.sac_agent import BehaviorSACAgent
from bark_ml.library_wrappers.lib_tf_agents.agents.ppo_agent import BehaviorPPOAgent
from bark_ml.library_wrappers.lib_tf_agents.agents.graph_sac_agent import BehaviorGraphSACAgent
from bark_ml.library_wrappers.lib_tf_agents.agents.graph_ppo_agent import BehaviorGraphPPOAgent


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import unittest
import numpy as np
import os
import gym
import matplotlib
# matplotlib.use('PS')
import time
import pprint

from bark.core.models.behavior import BehaviorConstantAcceleration
from bark.runtime.commons.parameters import ParameterServer
from bark_ml.environments.blueprints import ContinuousHighwayBlueprint, \
  DiscreteHighwayBlueprint, ContinuousMergingBlueprint, DiscreteMergingBlueprint, \
  ConfigurableScenarioBlueprint
from bark_ml.environments.single_agent_runtime import SingleAgentRuntime
from bark_ml.environments.counterfactual_runtime import CounterfactualRuntime
from bark_ml.library_wrappers.lib_tf_agents.agents.sac_agent import BehaviorSACAgent
import bark_ml.environments.gym
from bark_ml.commons.py_spaces import BoundedContinuous
from bark_ml.behaviors.cont_behavior import BehaviorContinuousML

class PyEnvironmentTests(unittest.TestCase):
  def test_envs_cont_rl(self):
    params = ParameterServer()
    cont_blueprints = []
    cont_blueprints.append(ContinuousHighwayBlueprint(params))
    cont_blueprints.append(ContinuousMergingBlueprint(params))

    for bp in cont_blueprints:
      env = SingleAgentRuntime(blueprint=bp, render=False)
      env.reset()
      for _ in range(0, 10):
        action = np.random.uniform(low=-0.1, high=0.1, size=(2, ))
        observed_next_state, reward, done, info = env.step(action)
        # print(f"Reward: {reward}, Done: {done}")

  def test_envs_discrete_rl(self):
    params = ParameterServer()
    discrete_blueprints = []
    discrete_blueprints.append(DiscreteHighwayBlueprint(params))
    discrete_blueprints.append(DiscreteMergingBlueprint(params))

    for bp in discrete_blueprints:
      env = SingleAgentRuntime(blueprint=bp, render=False)
      env.reset()
      for _ in range(0, 10):
        action = np.random.randint(low=0, high=3)
        observed_next_state, reward, done, info = env.step(action)
        # print(f"Reward: {reward}, Done: {done}")

  def test_gym_wrapping(self):
    # highway-v0: continuous
    # highway-v1: discrete
    # merging-v0: continuous
    # merging-v1: discrete
    # are registered here: import bark_ml.environments.gym

    cont_envs = [gym.make("highway-v0"), gym.make("merging-v0")]
    for env in cont_envs:
      env.reset()
      for _ in range(0, 10):
        action = np.random.uniform(low=-0.1, high=0.1, size=(2, ))
        observed_next_state, reward, done, info = env.step(action)
        print(f"Observed state: {observed_next_state}, Reward: {reward}, Done: {done}")

    cont_envs = [gym.make("highway-v1"), gym.make("merging-v1")]
    for env in cont_envs:
      env.reset()
      for _ in range(0, 10):
        action = np.random.randint(low=0, high=3)
        observed_next_state, reward, done, info = env.step(action)
        print(f"Observed state: {observed_next_state}, Reward: {reward}, Done: {done}")

  def test_counterfactual_runtime(self):
    params = ParameterServer()
    bp = ContinuousMergingBlueprint(params)
    # BehaviorConstantAcceleration::ConstAcceleration
    behavior_model_pool = []
    count = 0
    for a in [-3., 0., 3.]:
      local_params = params.AddChild("local_"+str(count))
      local_params["BehaviorConstantAcceleration"]["ConstAcceleration"] = a
      behavior = BehaviorConstantAcceleration(local_params)
      behavior_model_pool.append(behavior)
      count += 1
    
    env = CounterfactualRuntime(
      blueprint=bp,
      render=True,
      behavior_model_pool=behavior_model_pool,
      params=params)
    sac_agent = BehaviorSACAgent(environment=env,
                                 params=params)
    env.ml_behavior = sac_agent
    env.ml_behavior.set_action_externally = False
    env.reset()
    for _ in range(0, 10):
      action = np.random.uniform(low=-0.1, high=0.1, size=(2, ))
      observed_next_state, reward, done, info = env.step(action)
      print(f"Observed state: {observed_next_state}, Reward: {reward}, Done: {done}")
    pprint.pprint(env.tracer._states)

  def test_configurable_blueprint(self):
    params = ParameterServer(filename="bark_ml/tests/data/highway_merge_configurable.json")
    # continuous model
    ml_behavior = BehaviorContinuousML(params=params)
    bp = ConfigurableScenarioBlueprint(
      params=params,
      ml_behavior=ml_behavior)
    env = SingleAgentRuntime(blueprint=bp, render=False)
    # agent
    sac_agent = BehaviorSACAgent(environment=env,
                                 params=params)
    env.ml_behavior = sac_agent
    # test run
    env.reset()
    for _ in range(0, 10):
      action = np.random.randint(low=0, high=3)
      observed_next_state, reward, done, info = env.step(action)
  

if __name__ == '__main__':
  unittest.main()


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import numpy as np
import tensorflow as tf

# tfa specific
from tf_agents.environments import py_environment
from tf_agents.specs import array_spec
from tf_agents.trajectories import time_step as ts
from collections import OrderedDict

tf.compat.v1.enable_v2_behavior()


class PyBARKEnvironment(py_environment.PyEnvironment):
  """Wrapper for TensorFlow Agents (https://github.com/tensorflow/agents)

  Arguments:
    py_environment -- Base class for environment from tf_agents
  """
  def __init__(self, env):
    self._env = env
    self._action_spec = array_spec.BoundedArraySpec(
      shape=self._env.action_space.shape,
      dtype=np.float32,
      minimum=self._env.action_space.low,
      maximum=self._env.action_space.high,
      name='action')
    self._observation_spec = array_spec.BoundedArraySpec(
      shape=self._env.observation_space.shape,
      dtype=np.float32,
      minimum=self._env.observation_space.low,
      maximum=self._env.observation_space.high,
      name='observation')
    self._state = np.zeros(shape=self._env.observation_space.shape,
      dtype=np.float32)
    self._episode_ended = False

  def action_spec(self):
    return self._action_spec

  def observation_spec(self):
    return self._observation_spec

  def render(self):
    """Renders the enviornment
    """
    return self._env.render()

  def _reset(self):
    """Resets the wrapper
    """
    self._state = np.array(self._env.reset(), dtype=np.float32)
    self._episode_ended = False
    return ts.restart(self._state)

  def _step(self, action):
    """Steps the world for a given dt
    """
    if self._episode_ended:
      return self.reset()
    state, reward, self._episode_ended, _ = self._env.step(action)
    self._state = np.array(state, dtype=np.float32)
    if self._episode_ended:
      return ts.termination(self._state, reward=reward)
    else:
      return ts.transition(self._state, reward=reward, discount=0.99)


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.
import matplotlib
from matplotlib.patches import Polygon
from matplotlib import cm, colors
import matplotlib.pyplot as plt

import math

import numpy as np

from bark.core.viewer import *
from bark.core.geometry import *
from bark.core.models.dynamic import StateDefinition

from bark.runtime.viewer.viewer import BaseViewer


class MPViewer(BaseViewer):
    # we do not need an init function as pybind11 implements it
    def __init__(self, params=None, **kwargs):
        super(MPViewer, self).__init__(params=params, **kwargs)
        if 'axis' in kwargs:
          self.axes = kwargs.pop("axis")
        else:
          self.axes = plt.subplots()[1]
          # removes whitespace
          # plt.subplots_adjust(bottom=0.0, left=0.0, right=1.0, top=1)
        self._cmap = self.setupColormap()

    def drawPoint2d(self, point2d, color, alpha):
        self.axes.plot(
            point2d.x(),
            point2d.y(),
            color=self.getColor(color),
            alpha=alpha,
            marker='x')

    def drawLine2d(self, line2d, color='blue', alpha=1.0, dashed=False, zorder=1, **kwargs):
        line_width = kwargs.pop("linewidth", 1)
        marker = kwargs.pop("marker", None)
        marker_size = kwargs.pop("markersize", 12)
        linestyle = (0, (5, 10)) if dashed else 'solid'
        line2d_np = line2d.ToArray()
        self.axes.plot(
            line2d_np[:, 0],
            line2d_np[:, 1],
            linestyle=linestyle,
            color=self.getColor(color),
            alpha=alpha,
            zorder=zorder,
            linewidth=line_width,
            marker=marker,
            markersize = marker_size)

    def drawPolygon2d(self, polygon, color, alpha, facecolor=None, linewidth=1, zorder=10, hatch=''):
        points = polygon.ToArray()
        polygon_draw = matplotlib.patches.Polygon(
            points,
            True,
            facecolor=self.getColor(facecolor),
            edgecolor=self.getColor(color),
            alpha=alpha,
            linewidth=linewidth,
            zorder=zorder,
            hatch=hatch)
        t_start = self.axes.transData
        polygon_draw.set_transform(t_start)
        self.axes.add_patch(polygon_draw)
        center = polygon.center
        self.axes.plot(center[0], center[1], color=self.getColor(color))

    def drawTrajectory(self, trajectory, color='black', **kwargs):
        line_width = kwargs.pop("linewidth", 2)
        line_style = kwargs.pop("linestyle", "-")
        marker = kwargs.pop("marker", None)
        marker_size = kwargs.pop("markersize", 12)
        if len(trajectory) > 0:
            self.axes.plot(
                trajectory[:, int(StateDefinition.X_POSITION)],
                trajectory[:, int(StateDefinition.Y_POSITION)],
                color=self.getColor(color),
                linewidth=line_width,
                linestyle=line_style,
                marker=marker,
                markersize = marker_size)

    def drawArrow(self, pose):
        plt.annotate(s='', xy=(pose[0]+3*math.cos(pose[2]),pose[1]+3*math.sin(pose[2])), xytext=(pose[0],pose[1]), arrowprops=dict(arrowstyle='->'))

    def drawText(self, position, text, coordinate="axes", zorder=11, **kwargs):
        verticalalignment = kwargs.pop("verticalalignment", "top")
        horizontalalignment = kwargs.pop("horizontalalignment", "center")
        zorder = kwargs.pop("zorder", 12)
        t = None
        if coordinate=="axes":
            t = self.axes.text(position[0], position[1], text, horizontalalignment=horizontalalignment,
             verticalalignment=verticalalignment, transform=self.axes.transAxes, zorder=zorder, **kwargs)
        else:
            t = self.axes.text(position[0], position[1], text, horizontalalignment='center',
             verticalalignment='top', zorder=zorder, **kwargs)
        return t


    def drawCircle(self, position, radius):
        circle = plt.Circle(position, radius, color='r')
        self.axes.add_artist(circle)

    def getColor(self, color):
        if isinstance(color, Viewer.Color):
            return {
                Viewer.Color.White: 'w',
                Viewer.Color.Red: 'red',
                Viewer.Color.Blue: 'darkblue',
                Viewer.Color.Magenta: 'darkgreen',
                Viewer.Color.Brown: 'sienna'
            }.get(color, 'k')
        else:
            return color

    def get_aspect_ratio(self):
      # ratio is width / height
      [w, h] = self._get_ax_size()
      return (w/h)

    def _get_ax_size(self):
      fig = self.axes.get_figure()
      scale = fig.dpi_scale_trans.inverted()
      bbox = self.axes.get_window_extent().transformed(scale)
      width, height = bbox.width, bbox.height
      width *= fig.dpi
      height *= fig.dpi
      return width, height

    def drawWorld(self, world, eval_agent_ids=None, filename=None, scenario_idx=None, debug_text=True, axes_visible=False):
        self.clear()
        if not axes_visible:
            self.axes.set_axis_off()
        super(MPViewer, self).drawWorld(world, eval_agent_ids, filename, scenario_idx, debug_text)
        self._set_visualization_options(axes_visible)
        self.show()
        if filename:
            self.axes.get_figure().savefig(filename)

    def saveFig(self, filename):
      self.axes.get_figure().savefig(filename)

    def show(self, block=False):
        plt.draw()
        if block:
            plt.show(block=True)
        else:
            plt.pause(0.001)

    def _set_visualization_options(self, axes_visible):
        # x and y limits
        self.axes.set_xlim(self.dynamic_world_x_range[0], self.dynamic_world_x_range[1])
        self.axes.set_ylim(self.dynamic_world_y_range[0], self.dynamic_world_y_range[1])
        self.axes.get_xaxis().set_visible(axes_visible)
        self.axes.get_yaxis().set_visible(axes_visible) 

    def clear(self):
        self.axes.cla()

    def getSizeOfColormap(self):
        return 20 # cm.tab20 has 20 entries

    def setupColormap(self):
        cmap_np = cm.tab20(np.linspace(0, 1, 20)) # tab20 has 20 colours
        cmap_np = np.delete(cmap_np, [6,7], 0) # remove red (reserved for ego)
        return colors.ListedColormap(cmap_np)

    def getColorFromMap(self, double_color):
        return self._cmap(double_color)



from bark_ml.environments.blueprints.blueprint import Blueprint
from bark_ml.environments.blueprints.highway.highway import ContinuousHighwayBlueprint
from bark_ml.environments.blueprints.highway.highway import DiscreteHighwayBlueprint
from bark_ml.environments.blueprints.merging.merging import ContinuousMergingBlueprint
from bark_ml.environments.blueprints.merging.merging import DiscreteMergingBlueprint
from bark_ml.environments.blueprints.intersection.intersection import ContinuousIntersectionBlueprint
from bark_ml.environments.blueprints.intersection.intersection import DiscreteIntersectionBlueprint
from bark_ml.environments.blueprints.configurable.configurable_scenario import ConfigurableScenarioBlueprint


# Copyright (c) 2020 fortiss GmbH
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


class AgentTrackInfo:
    def __init__(self, filename, track_id, start_time, end_time, first_ts_on_map=None, precision=-2):
        self._filename = filename
        self._track_id = track_id  # TODO: rename to agent_id
        self._start_time = int(round(start_time, precision))  # in ms
        self._end_time = int(round(end_time, precision))  # in ms

    def __str__(self):
        return 'id={} start_time={} end_time={}'.format(self._track_id, self._start_time, self._end_time)

    def GetFileName(self):
        # yields track filename
        return self._filename

    def GetTrackId(self):
        return self._track_id

    def GetStartTimeMs(self):
        return self._start_time

    def GetEndTimeMs(self):
        return self._end_time



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

from bark.runtime.scenario import Scenario
from bark.runtime.scenario.scenario_generation import ScenarioGeneration
from bark.runtime.scenario.interaction_dataset_processing import InteractionDatasetReader
from bark.runtime.scenario.interaction_dataset_processing import DatasetDecomposer
from bark.runtime.commons.model_json_conversion import ModelJsonConversion
from bark.runtime.commons import ParameterServer
# PyBind imports
from bark.core.world.map import *
from bark.core.models.dynamic import *
from bark.core.models.execution import *
from bark.core.world.opendrive import *
from bark.core.world.goal_definition import *
from bark.core.geometry import *

import logging


class InteractionDatasetScenarioGenerationFull(ScenarioGeneration):
    # This class reads in a track file from the interaction dataset
    # and generates a scenario for each agent as the eval agent.

    def __init__(self, params=None, num_scenarios=None, random_seed=None):
        self._map_interface = None
        super().__init__(params, num_scenarios, random_seed)

    def initialize_params(self, params):
        super().initialize_params(params)
        params_temp = \
            self._params["Scenario"]["Generation"]["InteractionDatasetScenarioGenerationFull"]
        self._map_file_name = params_temp["MapFilename",
                                          "Path to the open drive map",
                                          "bark/runtime/tests/data/DR_DEU_Merging_MT_v01_shifted.xodr"]
        self._track_file_name_list = params_temp["TrackFilenameList",
                                                 "List of Paths to track files (csv)",
                                                 ["bark/runtime/tests/data/interaction_dataset_DEU_Merging_dummy_track.csv"]]
        self._behavior_model = params_temp["BehaviorModel",
                                           "Overwrite static trajectory of others with behavior model", None]
        self._xy_offset = params_temp["XYOffset",
                                      "offset in x and y direction.", [0, 0]]
        self._excluded_tracks = params_temp[
            "ExcludeTracks", "Track IDs to be excluded from the scenario generation", []]
        self._base_params_json = params_temp[
            "BaseParams", "Initial parameters of each scenario", ParameterServer(log_if_default=True)].ConvertToDict()
        self._agent_params = []
        self._starting_offset_ms = params_temp["StartingOffsetMs",
                                               "Starting Offset to each agent in miliseconds", 500]
        self._road_ids = params_temp["RoadIds",
                                     "Road ids for road corridor.", [0, 1]]
        self._vehicle_length_max = params_temp["VehicleLengthMax",
                                               "Maximum allowed vehicle length", 5.0]
        self._use_shape_from_track = params_temp["UseShapeFromTrack",
                                                 "Use shape from track", True]
        self._use_rectangle_shape = params_temp["RectangleShape",
                                                "Use Rectangle vehicle shape", True]
        self._use_goal_from_road = params_temp["GoalFromRoad",
                                               "Use goal from road", False]
        self._rel_pose_goal_on_road = params_temp["RelPoseGoalOnRoad",
                                                  "Relative position of goal on road", 0.99]

        self._interaction_ds_reader = InteractionDatasetReader(
            use_shape_from_track=self._use_shape_from_track, use_rectangle_shape=self._use_rectangle_shape)

    # TODO: remove code duplication with configurable scenario generation
    def create_scenarios(self, params, num_scenarios):
        """
            see baseclass
        """
        scenario_list = []

        self._map_interface = self.__create_map_interface__()
        self._road_corridor = self.__create_road_corridor__()

        for track_file_name in self._track_file_name_list:

            dataset_decomposer = DatasetDecomposer(map_interface=self._map_interface,
                                                   road_corridor=self._road_corridor,
                                                   track_filename=track_file_name,
                                                   vehicle_length_max=self._vehicle_length_max,
                                                   xy_offset=self._xy_offset,
                                                   starting_offset_ms=self._starting_offset_ms)
            scenario_track_info_list = dataset_decomposer.decompose()

            num = min(num_scenarios, len(scenario_track_info_list))

            for idx_s, sti in enumerate(scenario_track_info_list):
                if idx_s < num_scenarios and sti.GetEgoTrackInfo().GetTrackId() not in self._excluded_tracks:

                    logging.info("Creating scenario {}/{}".format(idx_s, num))
                    try:
                        scenario = self.__create_single_scenario__(sti)
                    except:
                        raise ValueError(
                            "Generation of scenario failed: {}".format(sti))
                    scenario_list.append(scenario)
                else:
                    break
        return scenario_list

    def __create_single_scenario__(self, scen_track_info):
        scen_track_info.TimeSanityCheck()

        scenario = Scenario(map_file_name=self._map_file_name,
                            json_params=self._params.ConvertToDict(),
                            map_interface=self._map_interface)

        world = scenario.GetWorldState()
        track_params = ParameterServer()
        track_params["execution_model"] = 'ExecutionModelInterpolate'
        track_params["dynamic_model"] = 'SingleTrackModel'
        track_params["map_interface"] = world.map

        all_track_ids = list(scen_track_info.GetOtherTrackInfos().keys())
        # also add ego id
        ego_track_id = scen_track_info.GetEgoTrackInfo().GetTrackId()
        all_track_ids.append(ego_track_id)

        agent_list = []
        model_converter = ModelJsonConversion()
        for track_id in all_track_ids:
            if self._behavior_model and track_id != ego_track_id:
                behavior_params = self.__fill_agent_params()
                behavior_model_name = self._behavior_model
                track_params["behavior_model"] = model_converter.convert_model(
                    behavior_model_name, behavior_params)
                # behavior_params.Save("/tmp/agent_prams_{}.json".format(track_id))
            else:
                # we do not change behavior model of ego agent -> will be set in benchmark
                track_params["behavior_model"] = None

            if self._use_goal_from_road:
                goal = self.__infer_goal_from_road__(lc=0)
            else:
                goal = None

            agent = self._interaction_ds_reader.AgentFromTrackfile(
                track_params, self._params, scen_track_info, track_id, goal_def=goal)

            # set first valid time stamp of the agent (in relation to scenario start)
            agent.first_valid_timestamp = scen_track_info.GetTimeOffsetOfAgentInSec(
                track_id)

            agent_list.append(agent)

        scenario._agent_list = agent_list  # must contain all agents!
        scenario._eval_agent_ids = [ego_track_id]
        scenario.json_params["track_file"] = scen_track_info.GetTrackFilename()

        return scenario

    def __fill_agent_params(self):
        agent_params = ParameterServer(
            log_if_default=True, json=self._base_params_json)
        self._agent_params.append(agent_params)
        # print("\n", agent_params.ConvertToDict())
        return agent_params

    def __create_map_interface__(self):
        params = ParameterServer()
        # we are creating a dummy scenario to get the map interface from it
        scenario = Scenario(map_file_name=self._map_file_name,
                            json_params=params.ConvertToDict())
        world = scenario.GetWorldState()
        map_interface = world.map

        return map_interface

    def __create_road_corridor__(self):
        if self._map_interface is not None:
            map_interface = self._map_interface
            map_interface.GenerateRoadCorridor(
                self._road_ids, XodrDrivingDirection.forward)
        else:
            raise ValueError("map interface not available")
        return map_interface.GetRoadCorridor(self._road_ids, XodrDrivingDirection.forward)

    def __infer_goal_from_road__(self, lc):
        lane_corr = self._road_corridor.lane_corridors[0]
        goal_polygon = Polygon2d([0, 0, 0], [
                                 Point2d(-10, -10), Point2d(-10, 10), Point2d(10, 10), Point2d(10, -10)])
        goal_point = GetPointAtS(
            lane_corr.center_line, lane_corr.center_line.Length()*self._rel_pose_goal_on_road)
        goal_polygon = goal_polygon.Translate(goal_point)
        return GoalDefinitionPolygon(goal_polygon)



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import unittest
import numpy as np
import os
import gym
import matplotlib
import time

# BARK imports
from bark.runtime.commons.parameters import ParameterServer

# BARK-ML imports
from bark_ml.environments.blueprints import ContinuousHighwayBlueprint, \
  DiscreteHighwayBlueprint, ContinuousMergingBlueprint, DiscreteMergingBlueprint
from bark_ml.environments.single_agent_runtime import SingleAgentRuntime
import bark_ml.environments.gym
from bark_ml.library_wrappers.lib_tf_agents.agents.ppo_agent import BehaviorPPOAgent
from bark_ml.library_wrappers.lib_tf_agents.agents.sac_agent import BehaviorSACAgent
from bark_ml.library_wrappers.lib_tf_agents.runners.ppo_runner import PPORunner
from bark_ml.library_wrappers.lib_tf_agents.runners.sac_runner import SACRunner


class PyLibraryWrappersTFAgentTests(unittest.TestCase):
  # make sure the agent works
  def test_agent_wrapping(self):
    params = ParameterServer()
    env = gym.make("highway-v0")
    env.reset()
    agent = BehaviorPPOAgent(environment=env,
                     params=params)
    agent = BehaviorSACAgent(environment=env,
                     params=params)

  # assign as behavior model (to check if trained agent can be used)
  def test_behavior_wrapping(self):
    # create scenario
    params = ParameterServer()
    bp = ContinuousHighwayBlueprint(params,
                                    num_scenarios=10,
                                    random_seed=0)
    env = SingleAgentRuntime(blueprint=bp,
                             render=False)
    ml_behaviors = []
    ml_behaviors.append(
      BehaviorPPOAgent(environment=env,
                       params=params))
    ml_behaviors.append(
      BehaviorSACAgent(environment=env,
                       params=params))
    
    for ml_behavior in ml_behaviors:
      # set agent
      env.ml_behavior = ml_behavior
      env.reset()
      done = False
      while done is False:
        action = np.random.uniform(low=-0.1, high=0.1, size=(2, ))
        observed_next_state, reward, done, info = env.step(action)
        print(f"Observed state: {observed_next_state}, Reward: {reward}, Done: {done}")
      
      # action is set externally
      ml_behavior._set_action_externally = True
      agent_id = list(env._world.agents.keys())[0]
      observed_world = env._world.Observe([agent_id])[0]
      action = np.random.uniform(low=-0.1, high=0.1, size=(2, ))
      ml_behavior.ActionToBehavior(action)
      a = ml_behavior.Plan(0.2, observed_world)
      action = np.random.uniform(low=-0.1, high=0.1, size=(2, ))
      ml_behavior.ActionToBehavior(action)
      b = ml_behavior.Plan(0.2, observed_world)
      self.assertEqual(np.any(np.not_equal(a, b)), True)

      # action will be calculated within the Plan(..) fct.
      a = ml_behavior.Plan(0.2, observed_world)
      b = ml_behavior.Plan(0.2, observed_world)
      np.testing.assert_array_equal(a, b)


  # agent + runner
  def test_agent_and_runner(self):
    params = ParameterServer()
    bp = ContinuousHighwayBlueprint(params,
                                    num_scenarios=10,
                                    random_seed=0)
    env = SingleAgentRuntime(blueprint=bp, render=False)
    agent = BehaviorPPOAgent(environment=env, params=params)
    
    # set agent
    params["ML"]["PPORunner"]["NumberOfCollections"] = 2
    params["ML"]["SACRunner"]["NumberOfCollections"] = 2
    params["ML"]["TFARunner"]["EvaluationSteps"] = 2
    env.ml_behavior = agent
    self.assertEqual(env.ml_behavior.set_action_externally, False)
    runner = PPORunner(params=params,
                       environment=env,
                       agent=agent)
    runner.Train()
    self.assertEqual(env.ml_behavior.set_action_externally, True)
    runner.Run()
    self.assertEqual(env.ml_behavior.set_action_externally, True)



if __name__ == '__main__':
  unittest.main()


#!/usr/bin/env python

import sys


def get_value_list(d):
    assert isinstance(d, dict)
    if sys.version_info[0] == 2:
        item_list = d.values()
    elif sys.version_info[0] == 3:
        item_list = list(d.values())
    else:
        # should not happen
        raise RuntimeError("Only python 2 and 3 supported.")
    assert isinstance(item_list, list)
    return item_list


def get_item_iterator(d):
    assert isinstance(d, dict)
    if sys.version_info[0] == 2:
        item_iter = d.iteritems()
        assert hasattr(item_iter, "next")
    elif sys.version_info[0] == 3:
        item_iter = iter(d.items())
        assert hasattr(item_iter, "__next__")
    else:
        # should not happen
        raise RuntimeError("Only python 2 and 3 supported.")
    assert hasattr(item_iter, "__iter__")
    return item_iter



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import logging

import numpy as np
import pandas as pd

from bark.core.world.opendrive import *
from bark.core.world import *
from bark.core.geometry import *
from bark.core.runtime import PyRuntime
from bark.core.models.dynamic import StateDefinition


class Runtime(PyRuntime):
  def __init__(self,
               step_time,
               viewer,
               scenario_generator=None,
               render=False,
               maintain_world_history=False):
    self._step_time = step_time
    self._viewer = viewer
    self._scenario_generator = scenario_generator
    self._scenario_idx = None
    self._scenario = None
    self._render = render
    self._reset_has_been_called = False
    self._maintain_world_history = maintain_world_history
    self._world_history = []

  def reset(self, scenario=None):
    if scenario:
      self._scenario = scenario
    else:
      self._scenario, self._scenario_idx = \
        self._scenario_generator.get_next_scenario()
    self._world = self._scenario.GetWorldState()
    self._reset_has_been_called = True
    self._viewer.Reset()
    self._world_history = []

  def step(self):
    assert(self._reset_has_been_called == True)

    # save history at start of simulation
    if len(self._world_history) == 0 and self._maintain_world_history:
        self._world_history.append(self._world.Copy())

    self._world.Step(self._step_time)

    if self._render:
        self.render()
    
    if self._maintain_world_history:
      self._world_history.append(self._world.Copy())

  def render(self):
    # self._viewer.clear()
    self._viewer.drawWorld(
      self._world,
      self._scenario._eval_agent_ids,
      scenario_idx=self._scenario_idx)
    self._viewer.clear()

  def run(self, steps):
    for step_count in range(steps):
      self.Step()

  def ExtractTimeSeries(self):
    # world history needs to be saved during simulation!
    state_list = []
    action_list = []
    for w in self._world_history:
        eval_agent = w.GetAgent(self._scenario._eval_agent_ids[0])
        state_list.append(eval_agent.state)
        action_list.append(eval_agent.behavior_model.GetLastAction())
    traj = np.stack(state_list, axis=0)
    act_arr = np.stack(action_list, axis=0)

    d = {'time': traj[:, int(StateDefinition.TIME_POSITION)],
          'x': traj[:, int(StateDefinition.X_POSITION)],
          'y': traj[:, int(StateDefinition.Y_POSITION)],
          'theta': traj[:, int(StateDefinition.THETA_POSITION)],
          'v': traj[:, int(StateDefinition.VEL_POSITION)]}

    # makes this generic for all sorts of input types
    if not np.isscalar(eval_agent.behavior_model.GetLastAction()):
        d["a"] = act_arr[:, 0]
        d["delta"] = act_arr[:, 1]
        if act_arr.shape[1] > 2:
            logging.warning("Incorrect parsing of actions to timeseries")
    else:
        d["acc"] = act_arr        
    df = pd.DataFrame(data=d)
    return df



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.



from bark.runtime.scenario.scenario_generation.config_readers.config_readers_interfaces import ConfigReaderDynamicModels

from bark.core.models.dynamic import *
from bark.runtime.commons.parameters import ParameterServer

  # this config reader defines dynamic models with fixed type for all agents
class FixedDynamicType(ConfigReaderDynamicModels):
  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    model_type = config_param_object["ModelType", "Type of dynamic model \
                used for all vehicles", "SingleTrackModel"]
    dynamic_models = []
    for _ in agent_states:
      bark_model = self.model_from_model_type(model_type)
      dynamic_models.append(bark_model)
    return dynamic_models, {}, config_param_object

  def default_params(self):
    return {"ModelType": "SingleTrackModel"}

  def model_from_model_type(self, model_type):
    params = ParameterServer()
    bark_model = eval("{}(params)".format(model_type))    
    return bark_model



# Copyright (c) 2020 Patrick Hart, Julian Bernhard,
# Klemens Esterle, Tobias Kessler
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT
import numpy as np
# BARK
from bark.core.world.evaluation import \
  EvaluatorGoalReached, EvaluatorCollisionEgoAgent, \
  EvaluatorStepCount, EvaluatorDrivableArea, CaptureAgentStates
from bark.runtime.commons.parameters import ParameterServer
from bark.core.geometry import *
# BARK-ML
from bark_ml.evaluators.evaluator import StateEvaluator


class GoalReachedGuiding(StateEvaluator):
  def __init__(self,
               params=ParameterServer(),
               eval_agent=None):
    StateEvaluator.__init__(self, params)
    self._goal_reward = \
      self._params["ML"]["GoalReachedGuiding"]["GoalReward",
        "Reward for reaching the goal.",
        1.]
    self._col_penalty = \
      self._params["ML"]["GoalReachedGuiding"]["CollisionPenalty",
        "Reward given for a collisions.",
        -1.]
    self._max_steps = \
      self._params["ML"]["GoalReachedGuiding"]["MaxSteps",
        "Maximum steps per episode.",
        50]
    self._act_penalty = \
      self._params["ML"]["GoalReachedGuiding"]["ActionPenalty",
        "Weight factor for penalizing actions",
        0.001]
    self._goal_dist = \
      self._params["ML"]["GoalReachedGuiding"]["GoalDistance",
        "Weight factor for distance to goal",
        0.01]
    self._eval_agent = eval_agent
    self._goal_lane_corr = None

  def _add_evaluators(self):
    """Evaluators that will be set in the BARK world"""
    evaluators = {}
    evaluators["goal_reached"] = EvaluatorGoalReached()
    evaluators["collision"] = EvaluatorCollisionEgoAgent()
    evaluators["step_count"] = EvaluatorStepCount()
    evaluators["drivable_area"] = EvaluatorDrivableArea()
    return evaluators

  def GetGoalLaneCorridorForGoal(self, observed_world):
    """Returns the lanecorridor the goal is in"""
    if self._goal_lane_corr is not None:
      return self._goal_lane_corr
    ego_agent = observed_world.ego_agent
    goal_def = ego_agent.goal_definition
    goal_shape = goal_def.goal_shape
    rc = observed_world.ego_agent.road_corridor
    lane_corr = None
    for lc in rc.lane_corridors:
      if Collide(lc.polygon, goal_shape):
        lane_corr = lc
    return lane_corr

  def CalculateDistanceToGoal(self, observed_world, goal_lane_corr):
    """Calculates the distance to the goal of the ego_agent"""
    goal_center_line = goal_lane_corr.center_line
    ego_agent = observed_world.ego_agent
    ego_agent_state = ego_agent.state
    distance_to_goal = Distance(
      goal_center_line,
      Point2d(ego_agent_state[1], ego_agent_state[2]))
    return distance_to_goal

  def CalculateGuidingReward(self, observed_world, action):
    """Returns a guiding reward using the dist. to goal and penalized acts."""
    guiding_reward =  0.
    goal_lane_corr = self.GetGoalLaneCorridorForGoal(observed_world)
    distance_to_goal = self.CalculateDistanceToGoal(observed_world, goal_lane_corr)
    guiding_reward -= self._goal_dist*distance_to_goal
    
    # NOTE: hack
    # desired_velocity = 10
    
    # NOTE: this will only work for cont. actions
    if action is not None and type(action) is not int:
      accs = action[0]
      delta = action[1]
      guiding_reward -= self._act_penalty*(accs**2 + delta*+2)
    return guiding_reward

  def _evaluate(self, observed_world, eval_results, action):
    """Returns information about the current world state
    """
    done = False
    ego_agent = observed_world.ego_agent
    ego_agent_state = ego_agent.state
    step_count = eval_results["step_count"]
    success = eval_results["goal_reached"]
    bad_state = eval_results["collision"] or eval_results["drivable_area"] \
      or (step_count > self._max_steps) or (ego_agent_state[4] < 0)
    # for agent_id, agent in observed_world.agents.items():
    #   eval_results[agent_id] = agent.state
    # determine whether the simulation should terminate
    if success or bad_state or step_count > self._max_steps:
      done = True
    if bad_state:
      success = 0
      eval_results["goal_reached"] = 0
    
    guiding_reward = self.CalculateGuidingReward(observed_world, action)
    # calculate reward
    reward = bad_state * self._col_penalty + \
      success * self._goal_reward + self._goal_dist*guiding_reward
    return reward, done, eval_results
    
  def Reset(self, world):
    self._goal_lane_corr = None
    return super(GoalReachedGuiding, self).Reset(world)


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import unittest
import numpy as np
import os
import matplotlib
import time

from bark_ml.behaviors.cont_behavior import BehaviorContinuousML
from bark_ml.behaviors.discrete_behavior import BehaviorDiscreteMotionPrimitivesML, \
        BehaviorDiscreteMacroActionsML
from bark.runtime.commons.parameters import ParameterServer
from bark.core.models.dynamic import SingleTrackModel
from bark.core.world import World, MakeTestWorldHighway


class PyBehaviorTests(unittest.TestCase):
  def test_discrete_behavior(self):
    params = ParameterServer()
    discrete_behavior = BehaviorDiscreteMacroActionsML(params)
    # sets 0-th motion primitive active
    discrete_behavior.ActionToBehavior(0)
    print(discrete_behavior.action_space)
    
  def test_cont_behavior(self):
    params = ParameterServer()
    cont_behavior = BehaviorContinuousML(params)
    # sets numpy array as next action
    cont_behavior.ActionToBehavior(np.array([0., 0.]))
    print(cont_behavior.action_space)


if __name__ == '__main__':
  unittest.main()